/*======================================================================================
  
  Tffind release version 1.1
  Coded by Matt Weirauch (weirauch@soe.ucsc.edu)
  Nov 3, 2003


  Find patterns of transcription factor binding sites in a pipmaker multiple alignment
  file.  The basic format of a pattern is:

	{GATA-1[0.85]: IMD} <10,200> {NF-kB(c)[0.85]: IMD}

  Upon being given the preceding pattern, Tffind proceeds to locate all occurences of
  the GATA-1 binding site followed by the NF-kB binding site 10 to 200 bps downstream and
  on the reverse strand.  Patterns may be arbitrarily large or small.

  For documentation and source code, see the the Multipipmaker tools section at
					 http://globin.cse.psu.edu/
  For command line options, type "tffind"

  Updates for version 1.1:
	- added '-lc' option, which treats lowercase nucleotides as if they were 
	  uppercase (i.e. doesn't mask them out)
	- added '-gff' option, which changes the output to gff format
	- now accepts spacers of format "0,100" as well as "<0,100>"

  ======================================================================================*/

#include "tffind.h"

int main( int argc, char *argv[] )
{
    char align_type,    /* type of alignment file (f, m or b) */
	 *align_file,   /* name of alignment file */
	 *mat_file,	/* name of matrix file */
	 *exon_file,    /* name of file containing exon and gene intervals */
	 *pat_string,	/* pattern created by user to search for */
	 **A = NULL, 	/* the chars of the alignment, used by type 'f' and 'm' files */
	 **seq_names;	/* the name of each seq in the alignment */
    int i,			/* array index */
	nrow, 			/* number of seqs (rows) in the alignment */
    	ncol, 			/* number of columns in the alignment */
	cur_dir, 		/* current strand being searched (FORWARD or REVERSE) */
	min_matches, 		/* minimum number of seqs that must match the pattern */
	from, to,		/* range in sequence to search for the pattern */
	tot_matches = 0, 	/* total matches found to the pattern */
	num_motifs = 0, 	/* number of motifs in the pattern */
	num_to_mask = 0, 	/* number of exon or gene regions to mask out */
	search_from, search_to, /* range used by each strand to search for the pattern */
	filter_exons = 0, 	/* filter out exon or gene regions ? (bool) */
	num_filtered_exons = 0, /* number of hits falling in exon regions filtered out */
	report_to_top,		/* report match in relation to topmost seq ? (bool) */
	exon_seq = 0,		/* which sequence exon positions are relative to */
	max_pat_len = 0,	/* maximum possible length the pattern can have */
	highest_pos = -1,	/* highest possible position in topmost sequence */
	count_lc = 0,		/* treat lowercase nucs equally to uppercase? */
	out_format = TFFIND_OUT,/* output format: TFFIND_OUT or GFF_OUT */
	*look_in_seq = NULL, 	/* look in this seq for the pattern ? (bool) */
	*beg = NULL, 		/* parallel arrays specifying beg of range, */
	*end = NULL, 		/*			      end of range, */
	*dir = NULL, 		/* 			      direction of exon */
	**col2pos = NULL;	/* gives pos in sequence given the col of alignment */
    struct ITEM *item_head = NULL;	/* linked list of motifs and spacers */
    struct MATCH *match_head = NULL;	/* linked list of matches found to pattern */
    Block_ptr block_head = NULL,        /* linked list of blocks in alignment (type b) */
              b_step = NULL;            /* pointer to step through linked list */
    Seq_ptr seq = NULL;			/* pointer to a seq in a set of contigs */
    Contig_ptr contig = NULL;		/* header, length, and string of nucleotides */
    SEQ *seq_fasta = NULL;		/* SEQ structure for files in FASTA format */

    /* allocate memory */
    align_file = ckalloc( STR_SIZE );
    mat_file = ckalloc( STR_SIZE );
    exon_file = ckalloc( STR_SIZE );
    pat_string = ckalloc( STR_SIZE );
    err_string = ckalloc( STR_SIZE );
    argv0 = ckalloc( sizeof( char ) * 7 );

    /* process command-line arguments */
    strcpy( argv0, "tffind");
    if (argc < 3)
	fatal( usage( SYNTAX ) );
    strcpy( align_file, argv[1]);
    strcpy( mat_file, argv[2]);
    align_type = determine_align_type( align_file );

    /* old multipipmaker output format */
    if (align_type == 'm')
    {
    	A = read_malign( align_file, &nrow, &ncol, &seq_names );

        /* determine mapping from alignment columns to positions in each seq */
	col2pos = (int **) ckalloc( sizeof( int *) * nrow );
	for (i = 0; i < nrow; i++)
	{
	    col2pos[i] = (int *) ckalloc( sizeof( int ) * ncol );
    	    get_position_in_seq( A, ncol, i, &(col2pos[i]) );
	}

	/* determine highest possible position in top sequence */
	for (i = ncol - 1; col2pos[0][i] == 0; i--);
        highest_pos = col2pos[0][i];

    /* new, 'block' multiple alignment format */
    } else if (align_type == 'b') {
	error( "Block alignment format not available for this release", "", KILL );
	block_head = read_blocks( align_file, &seq, &nrow, &contig, &i );
        ncol = contig->len;

 	/* initialize sequence names array */
        seq_names = (char **) ckalloc( sizeof( char *) * nrow );
	for (i = 0; i < nrow; i++)
	{
	    seq_names[i] = (char *) ckalloc( STR_SIZE );
	    strcpy(seq_names[i], seq[i].name);
	}

	/* initialize block numbers: each block is numbered in sequential order */
        /* also, determine highest possible position in top sequence */
	for (i = 1, b_step = block_head; b_step; b_step = b_step->next, i++)
	{
	    b_step->mark = i;
	    if (b_step->seg[0].contig_nbr == 0)
	        if (b_step->seg[0].end > highest_pos)
		    highest_pos = b_step->seg[0].end;
	}

    /* FASTA format - essentially treated as a special case of a type 'm' file
       with nrow = 1 */
    } else {
	seq_fasta = seq_get( align_file );
	ncol = seq_fasta->slen;
	nrow = 1;

	/* initialize sequence data */
	A = (char **) ckalloc( sizeof( char* ) * 1 );
	A[0] = (char *) ckalloc( sizeof( char ) * ncol );
	strncpy( A[0], seq_fasta->seq, seq_fasta->slen );

	/* initialize sequence names array */
	seq_names = (char **) ckalloc( sizeof( char* ) * 1);
	seq_names[0] = (char *) ckalloc( STR_SIZE );
	strncpy( seq_names[0], seq_fasta->header, seq_fasta->hlen );

	/* initialize col2pos array and highest_pos */
        col2pos = (int **) ckalloc( sizeof( int *) * 1 );
  	col2pos[0] = (int *) ckalloc( sizeof( int ) * ncol );
	for (i = 0; i < ncol; i++)
	    col2pos[0][i] = i + 1;
	highest_pos = ncol;

	/* close seq_fasta: we already have all of the info we need */
	seq_close( seq_fasta );
    }
    look_in_seq = ckalloc( sizeof( int ) * nrow );

    /* in interactive mode */
    if (argc == 3) {
        process_main_menu_choices( &pat_string, &item_head, &num_motifs, &from, &to,
			           align_file, look_in_seq, &filter_exons, nrow, ncol,
			           &min_matches, &report_to_top, &beg, &end, &dir,
				   &exon_seq, &num_to_mask, &max_pat_len, mat_file,
				   highest_pos, &exon_file, &count_lc, &out_format );

    /* in batch mode */
    } else {
        process_cmnd_line_options( &pat_string, &item_head, &num_motifs, &from, &to,
                                   look_in_seq, &filter_exons, nrow, ncol, &min_matches,
                                   &report_to_top, &beg, &end, &dir, &exon_seq,
				   &num_to_mask, argc, argv, &max_pat_len, mat_file, 
				   highest_pos, &exon_file, &count_lc, &out_format );
    }

    print_header( align_file, from, to, pat_string, nrow, ncol, filter_exons, exon_seq,
		  look_in_seq, min_matches, align_type, report_to_top, num_motifs,
		  exon_file, count_lc, out_format );

    /* initialize range of search */
    search_from = from;
    search_to = to;

    /* search for pattern on each strand */
    for (cur_dir = FORWARD; cur_dir >= REVERSE; cur_dir--)
    {
	fprintf(SCREEN, "\nSearching for pattern on %s strand...\n", 
		        cur_dir == FORWARD ? "forward" : "reverse");

        adjust_from_and_to( &search_from, &search_to, (col2pos ? col2pos[0] : NULL), 
			    ncol, cur_dir, align_type, max_pat_len );
        search_for_pattern( A, ncol, nrow, item_head, num_motifs, search_from, search_to,
			    cur_dir, look_in_seq, &match_head, align_type, block_head,
			    contig, col2pos, report_to_top, min_matches, count_lc );

	/* debug */ /* print_matches( match_head, num_motifs, nrow ); */

        output_matches( match_head, min_matches, filter_exons, &num_filtered_exons,
			&tot_matches, A, nrow, cur_dir, num_to_mask, beg, end, dir, from,
			to, align_type, block_head, contig, num_motifs, exon_seq,
			out_format );

        free_match_list( &match_head );
    }

    print_trailer( tot_matches, filter_exons, num_filtered_exons, nrow, seq_names,
		   report_to_top, align_type, out_format );

    /* free memory */
    free_item_list( &item_head );
    free_block_list( &block_head );
    free_int_arr( col2pos, nrow );
    free_char_arr( seq_names, nrow );
    free_contig( contig, nrow );
    free_seq( seq, nrow );
    ckfree( argv0 );
    ckfree( align_file );
    ckfree( mat_file );
    ckfree( exon_file );
    ckfree( pat_string );
    ckfree( err_string );
    ckfree( look_in_seq );
    ckfree( beg );
    ckfree( end );
    ckfree( dir );
    if (align_type == 'f')
    	free_char_arr( A, nrow );
    else if (align_type == 'm')
    {
    	ckfree( A[0] );
	ckfree( A );
    }

    /* pacify compiler (unused function warning) */
    print_matches( NULL, -1, -1 );

    return 0;
}


/* determine the type of alignment file. */
static char determine_align_type( char *align_file )
{
    char buf[BUFSIZE];	/* buffer to read in lines from file */
    int num;		/* a number in the file */
    FILE *fp = NULL;	/* file pointer */

    /* attempt to open alignment file */
    fp = ckopen( align_file, "r" );
    fgets( buf, BUFSIZE, fp );
    if (fclose( fp ) != 0)
    {
    	sprintf( err_string, "Could not close '%s'.", align_file );
        error( err_string, "", NOKILL );
    }

    /* determine type of alignment file */
    if (strncmp( buf, BBTAG, BBLEN ) == 0)
    {
	if(!( starts( buf + 5, "verbose" )))
	    error( "The provided alignment file is not verbose.", "", KILL );
	return 'b';
    }
    else if (sscanf( buf, "%d %d", &num, &num ) == 2)
	return 'm';
    else if (buf[0] == '>')
	return 'f';
    else
	error( "Unknown alignment file format.", "", KILL );

    return '?';			/* pacify compiler */
}


/* Print the main menu of options to the screen, along with the current settings. */
static void print_main_menu( char *pat_string, char *mat_file, int nrow, float cutoff,
			     int min_matches, int from, int to, int *look_in_seq,
			     char *align_file, int filter_exons, int ncol,
			     int report_to_top, int count_lc, int out_format )
{
    int i;	/* array index */

    fprintf(SCREEN, "\n\n" MENU_BORDER);
    fprintf(SCREEN, " 			   MAIN MENU\n");
    fprintf(SCREEN, "			   ---- ----\n\n");
    fprintf(SCREEN, "N...Add motif to current pattern by name\n");
    fprintf(SCREEN, "P...Add motif to current pattern by consensus pattern\n");
    fprintf(SCREEN, "A...Add motif to current pattern by access number\n");
    fprintf(SCREEN, "I...Add motif to current pattern by i.d. (TransFac only)\n");
    fprintf(SCREEN, "S...Add spacer to current pattern\n");
    fprintf(SCREEN, "D...Delete previous sub-pattern\n");
    fprintf(SCREEN, "M...Print main menu\n");
    fprintf(SCREEN, "Q...Quit program (without finding pattern)\n");
    fprintf(SCREEN, "F...FIND PATTERN\n\n");
    fprintf(SCREEN, "1...List available t.f. binding sites in current matrix database\n");
    fprintf(SCREEN, "2...Change current matrix database\n");
    fprintf(SCREEN, "3...Change minimum number of sequences that must match pattern\n");
    fprintf(SCREEN, "4...Change cutoff value for score function\n");
    fprintf(SCREEN, "5...Search for pattern only within a range of the alignment or ");
    fprintf(SCREEN, "sequence\n");
    fprintf(SCREEN, "6...Choose which sequences to search on in the alignment\n");
    fprintf(SCREEN, "7...Toggle exons option\n");
    fprintf(SCREEN, "8...Toggle which sequence positions are in relation to\n");
    fprintf(SCREEN, "9...Toggle between masking lowercase nucleotides\n");
    fprintf(SCREEN, "0...Toggle between output formats (TFFIND and GFF)\n\n\n");
    fprintf(SCREEN, "Current pattern : %s\n", pat_string);
    fprintf(SCREEN, "Current alignment file : %s\n", align_file);
    fprintf(SCREEN, "Current matrix database : %s\n", mat_file);
    fprintf(SCREEN, "Minimum number of sequences that must match : %d out of %d\n", 
		    min_matches, nrow);
    fprintf(SCREEN, "Cutoff value for score function: %.2f\n", cutoff);
    fprintf(SCREEN, "Range of search : ");
    fprintf(SCREEN, ((from == 1) && (to == ncol)) ? "Entire Sequence\n" : "%d to %d.\n", 
		     from, to);
    fprintf(SCREEN, "Search in sequences : " );
    for (i = 0; i < nrow; i++)
	if (look_in_seq[i])
	    fprintf(SCREEN, "  %d", i);
    fprintf(SCREEN, "\nFilter out regions overlapped by exons file? : %s\n",
                    filter_exons ? "yes" : "no");
    fprintf(SCREEN, "Mask out lowercase nucleotides? : %s\n",
		    count_lc ? "no" : "yes");
    fprintf(SCREEN, "Report sequences in relation to %s.\n",
		    report_to_top ? "the topmost sequence"
				  : "the sequence the match occurs on");
    fprintf(SCREEN, "Output format : %s\n",
		    out_format == TFFIND_OUT ? "TFFIND format"
				   	     : "GFF format");
    fprintf(SCREEN, MENU_BORDER);
    fprintf(SCREEN, CHOICE_PROMPT);
}


/* Get the main menu choice.  Only processes the first character entered. */
static char get_menu_choice()
{
    char c,		/* one char entered by user */
	 excess = ' ';	/* excess chars after the first one entered */

    c = getchar();
    while ((excess != '\n') && (c != '\n'))	/* skip eol and other excess chars */
	excess = getchar();
	
    return (c);
}


/* Create pattern linked list and set all parameters by parsing command-line args. */
static void process_cmnd_line_options( char **pat_string, struct ITEM **item_head,
                                       int *num_motifs, int *from, int *to,
                                       int *look_in_seq, int *filter_exons, int nrow,
                                       int ncol, int *min_matches, int *report_to_top,
				       int **beg, int **end, int **dir, int *exon_seq,
				       int *num_to_mask,  int argc, char **argv,
				       int *max_pat_len, char *current_mat_file,
				       int highest_pos, char **exon_file, int *count_lc,
				       int *out_format )
{
    char file_type = '?',       /* type of matrix file (i or t) */
         mask_type,             /* type of masking to be performed (g, c, e) */
	 *cur_spacer,		/* current spacer to be added to the pattern */
	 *buf;			/* buffer read in from file */
    int i, j, 			/* array indices */
	first, second,		/* spacer values */
	which_seq,		/* which seq we are on */
	type = '?';		/* type of motif identifier: name, pattern, id, ac */
    float cutoff;               /* cutoff value for scoring function */
    FILE *fp;                   /* file pointer */
    
    cur_spacer = ckalloc( STR_SIZE );
    buf = ckalloc( BUFSIZE );

    /*initialize default values */
    strcpy( *pat_string, " " );
    *from = 1;
    *to = ncol;
    *min_matches = nrow;
    for (i = 0; i < nrow; i++)
        look_in_seq[i] = 1;
    *report_to_top = 0;
    cutoff = DEFAULT_CUTOFF_VAL;

    /* open matrix file and determine what type it is */
    fp = ckopen( current_mat_file, "r" );
    fgets( buf, BUFSIZE, fp );
    file_type = guess_matrix_type( buf );
    if (file_type == '?')
	error( "Unknown matrix file type.", "", KILL );
    rewind( fp );

    /* begin list with "dummy" spacer value of <0,0> */
    add_spacer( 0, 0, pat_string, NULL, item_head, max_pat_len );

    /* iterate through command-line arguments */
    for (i = 3; i < argc; i++)
    {
	/* Add by name, pat, acnum, or id */
        if ((strcmp( argv[i], "-n" ) == 0) ||
	    (strcmp( argv[i], "-p" ) == 0) ||
	    (strcmp( argv[i], "-a" ) == 0) ||
	    (strcmp( argv[i], "-i" ) == 0)) {

	    /* determine type of add: name, pat, acnum, or id */
	    if (argv[i][1] == 'n')
	    	type = NAME;
	    else if (argv[i][1] == 'p')
		type = PAT;
            else if (argv[i][1] == 'a')
	   	type = AC;
            else if (argv[i][1] == 'i')
	    	type = ID;
	    else
		error( "Internal add_by error", "", KILL );

	    /* move to next argument */
            i++;
	    if (i == argc)
		fatal( usage( SYNTAX ) );

	    /* attempt to read info from database */
            if (!get_data_from_file( fp, argv[i], type, item_head, pat_string, cutoff,
                                     file_type, num_motifs, max_pat_len ))
	    {
		fprintf(SCREEN, "Not in database: %s\n", argv[i]);
		if (fclose( fp ) != 0)
		{
		    sprintf( err_string, "Could not close '%s'.", current_mat_file );
		    error( err_string, "", NOKILL );
	    	}
	 	exit(0);
	    }

	/* Add spacer */
        } else if (strcmp( argv[i], "-s" ) == 0) {
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );

	    strcpy( cur_spacer, argv[i] );
            if (!can_parse_spacer( &cur_spacer, &first, &second )) 
	    {
	  	fprintf(SCREEN, "Cannot parse spacer: %s\n", argv[i]);
                fprintf(SCREEN, "Spacer values must be non-negative,\nand the ");
                fprintf(SCREEN, "second value must be greater than the first.\n");
                fprintf(SCREEN, "Spacer format : <1,3> or <1>\n\n\n");
		exit(0);
            } else {
		if (*num_motifs != 0)
               	    add_spacer( first, second, pat_string, cur_spacer, item_head,
			    	    max_pat_len );
		else
		    error( "Patterns cannot have a spacer before the first motif.", "", 
			   KILL );
	    }

	/* List available transcription factors */
        } else if (strcmp( argv[i], "-list" ) == 0) {
            scan_matrix_file( fp, current_mat_file, file_type );

	/* Change matrix database */
        } else if (strcmp( argv[i], "-db" ) == 0) {
            i++;
	    if (i == argc)
                fatal( usage( SYNTAX ) );
            strcpy( current_mat_file, argv[i] );
            change_database( current_mat_file, &file_type, &fp, KILL );

	/* Change minimum number of seqs that must match */
        } else if (strcmp( argv[i], "-min" ) == 0) {
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );
            *min_matches = atoi( argv[i] );
            if (*min_matches <= 0 || *min_matches > nrow)
	    {
                sprintf( err_string, "Number of sequences that match must be positive"
                                     " and less than\n            the number of "
                                     "sequences in the alignment (%d).", nrow );
                error( err_string, "", KILL );
	    }

	/* Change cutoff value */
        } else if (strcmp( argv[i], "-cut" ) == 0) {
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );
            cutoff = atof( argv[i] );
            if (cutoff <= 0 || cutoff > 1)
            	error( "Cutoff must be positive and less than 1.", "", KILL ); 

	/* Specify range */
        } else if (strcmp( argv[i], "-range" ) == 0) {
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );
            *from = atoi( argv[i] );
    	    if ((*from <= 0) || (*from > highest_pos))
    	    {
        	sprintf( err_string, "Positions must be in the range 1 to %d.", 
			 highest_pos );
        	error( err_string, "", KILL );
	    }

            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );
            if (argv[i][0] == '.')
                *to = ncol;
            else
	    {
                *to = atoi( argv[i] );
            	if ((*to <= 0) || (*to > highest_pos) || (*to < *from))
            	{
                    sprintf( err_string, "Positions must be in the range 1 to %d.", 
			     highest_pos );
                    error( err_string, "", KILL );
            	}
	    }

	/* Specify which sequences to look in */
        } else if (strcmp( argv[i], "-lookin" ) == 0) {

            for (j = 0; j < nrow; j++)
                look_in_seq[j] = 0;

            i++;
            if (i == argc)
            	fatal( usage( SYNTAX ) );
	    which_seq = atoi( argv[i] );
	    
	    j = 0;
	    /* continue until we reach a non-integer */
	    while (!(which_seq == 0 && argv[i][0] != '0'))
	    {
            	if ((which_seq < 0) || (which_seq >= nrow))
	    	{
                    sprintf( err_string, "Sequence numbers must be between 0 and %d.",
                             (nrow - 1) );
                    error( err_string, "", KILL );
	    	}

		/* only count new sequences */
		if (look_in_seq[which_seq] == 0)
		    j++;

            	look_in_seq[which_seq] = 1;

            	i++;
            	if (i == argc)
		    break;

            	which_seq = atoi( argv[i] );
	    }

	    /* automatically update min_matches if its value is impossibly high */
	    if (*min_matches > j)
		*min_matches = j;

	    /* if we haven't reached the end, we have examined the next arg already and 
	       need to back up */
	    if (i != argc)
		i--;

	/* Filter out exon regions */
        } else if (strcmp( argv[i], "-exons" ) == 0) {
	    *filter_exons = 1;

	    /* first argument is type of mask: (g)ene, (c)oding region, (e)xon */
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );

	    mask_type = argv[i][0];
	    if ((mask_type != 'g') && (mask_type != 'c') && (mask_type != 'e'))
                fatal( usage( SYNTAX ) );

	    /* second argument is which sequence in alignment positions are relative to */
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );

    	    *exon_seq = atoi( argv[i] );
    	    if ((*exon_seq < 0) || (*exon_seq >= nrow))
	    {
        	sprintf( err_string, "Sequence must be between 0 and %d.", (nrow - 1) );
        	error( err_string, "", KILL );
	    }

	    /* third argument is name of exons file */
            i++;
            if (i == argc)
                fatal( usage( SYNTAX ) );

	    strcpy( *exon_file, argv[i] );
    	    fp = ckopen( *exon_file, "r" );
            *num_to_mask = get_exons( fp, beg, end, dir, mask_type );
            if (!(*num_to_mask))
	        error( "Invalid exons file format.", "", KILL );

	/* Specify which seq positions are reported in terms of */
        } else if (strcmp( argv[i], "-reltop" ) == 0) {
	    *report_to_top = 1;

	} else if (strcmp( argv[i], "-lc" ) == 0) {
	    *count_lc = 1;

	} else if (strcmp( argv[i], "-gff" ) == 0) {
	    *out_format = GFF_OUT;

	/* Invalid menu choice */
	} else {
            fatal( usage( SYNTAX ) );
	}
    }

    if (*num_motifs == 0)
	error( "No motifs have been entered.", "", KILL );

    if (fclose( fp ) != 0)
    {
    	sprintf( err_string, "Could not close '%s'.", current_mat_file );
	error( err_string, "", KILL );
    }

    remove_trailing_spacers( item_head, pat_string );
    ckfree( cur_spacer );
    ckfree( buf );
}


/* Create linked list that makes up the pattern entered by the user.
   Basically, set up everything for the search. */
static void process_main_menu_choices( char **pat_string, struct ITEM **item_head,
				       int *num_motifs, int *from, int *to,
				       char *align_file,int *look_in_seq,
				       int *filter_exons, int nrow, int ncol,
				       int *min_matches, int *report_to_top, int **beg,
				       int **end, int **dir, int *exon_seq,
			  	       int *num_to_mask, int *max_pat_len, 
				       char *current_mat_file, int highest_pos,
				       char **exon_file, int *count_lc, int *out_format )
{
    char choice, 		/* menu choice entered by user */
	 file_type = '?', 	/* type of matrix file (i or t) */
	 *long_string,		/* utility string for various user inputs, etc. */
	 *buf;			/* buffer read in from file */
    int i,			/* array index */ 
	first, second, 		/* spacer values, as entered by user */
	do_print_menu = 1, 	/* should main menu be reprinted on this iteration? */
	type = NAME;		/* how user is entering motif (NAME, PAT, AC, ID) */
    float cutoff;	 	/* cutoff value for scoring function */
    FILE *fp;			/* file pointer */

    /* allocate memory */
    long_string = ckalloc( STR_SIZE );
    buf = ckalloc( BUFSIZE );

    /* initialize default values */
    strcpy( *pat_string, " " );
    *from = 1;
    *to = ncol;
    *min_matches = nrow;
    for (i = 0; i < nrow; i++)
	look_in_seq[i] = 1;
    *report_to_top = 0;
    cutoff = DEFAULT_CUTOFF_VAL;

    /* attempt to open matrix file and determine its type */
    fp = ckopen( current_mat_file, "r" );
    fgets( buf, BUFSIZE, fp );
    file_type = guess_matrix_type( buf );
    if (file_type == '?')
	error( "Unknown matrix file type.", "", KILL );
    rewind( fp );

    /* begin list with "dummy" spacer value of <0,0> */
    add_spacer( 0, 0, pat_string, long_string, item_head, max_pat_len );

    while (1)
    {
   	if (do_print_menu)
            print_main_menu( *pat_string, current_mat_file, nrow, cutoff, *min_matches,
                             *from, *to, look_in_seq, align_file, *filter_exons, ncol,
			     *report_to_top, *count_lc, *out_format );
        do_print_menu = 1;
	choice = get_menu_choice();

	switch (choice)
	{
            case 'N' :          /* Add by name */
            case 'n' :

            case 'P' :          /* Add by consensus pattern */
            case 'p' :

	    case 'A' :		/* Add by access number */
	    case 'a' :

	    case 'I' :		/* Add by i.d. */
	    case 'i' :

		switch (choice)
		{
		    case 'N' : 
		    case 'n' : strcpy(long_string, "name of the motif"); type = NAME; 
			       break;

		    case 'P' :
		    case 'p' : strcpy(long_string, "consensus pattern motif"); type = PAT;
			       break;

		    case 'A' : 
	  	    case 'a' : strcpy(long_string, "access number"); type = AC;
			       break;

		    case 'I' :
		    case 'i' : strcpy(long_string, "TransFac i.d."); type = ID;
			       break;
		}

		if ((type == ID) && (file_type != 't'))
		{
		    error( "Non TransFac files do not have TransFac i.d.'s", "", NOKILL );
		    do_print_menu = 0;
                    fprintf(SCREEN, CHOICE_PROMPT);
		    break;
	 	}

                rewind( fp );
                fprintf(SCREEN, "\n\nEnter the %s that you would ", long_string);
                fprintf(SCREEN, "like to add.\nTo search on reverse strand, ");
                fprintf(SCREEN, "end it with (c).  ");
		get_entry (&long_string); 

                do_print_menu = get_data_from_file( fp, long_string, type, item_head,
                                                    pat_string, cutoff, file_type,
                                                    num_motifs, max_pat_len );
                if (!do_print_menu)
    		    fprintf(SCREEN, CHOICE_PROMPT);
		    
                break;

            case 'S' :          /* Add spacer */
            case 's' :

		if (*num_motifs == 0)
		{
                    error( "Patterns cannot have a spacer before the first motif.", "", 
			   NOKILL );
		    do_print_menu = 0;			
		    break;
		}

                fprintf(SCREEN, "\n\nEnter new spacer to be added.  ");
		get_entry (&long_string);  

                if (!can_parse_spacer( &long_string, &first, &second ))
                {
		    error( "Invalid spacer.", "", NOKILL );
		    fprintf(SCREEN, "Spacer values must be non-negative,\nand the ");
		    fprintf(SCREEN, "second value must be greater than the first.\n");
                    fprintf(SCREEN, "Spacer format : <1,3> or <1>\n\n\n");
    		    fprintf(SCREEN, CHOICE_PROMPT);
                    do_print_menu = 0;
                } else {
                    add_spacer( first, second, pat_string, long_string, item_head,
			        max_pat_len );
		}			
                break;

            case 'D' :          /* Delete last node */
            case 'd' :

                fprintf(SCREEN, "\n\nAre you sure that you want to delete the ");
                fprintf(SCREEN, "previous sub-pattern(y/n)? ");
                choice = get_menu_choice( );
                if ((choice == 'y') || (choice == 'Y'))
                {
                    delete_last_node( item_head, num_motifs );
                    delete_last_pat( pat_string );
                }
                break;

	    case 'M' :		/* Print Main Menu */
	    case 'm' :

		do_print_menu = 1;
		break;

            case 'Q' :          /* Quit */
            case 'q' :

		error( "Program terminated early by user. No output produced.", "", KILL);
                break;

            case 'F' :		/* Find pattern */
            case 'f' :

		if (*num_motifs == 0)
		{
		    error( "No motifs have been entered.", "", NOKILL );
    		    fprintf(SCREEN, CHOICE_PROMPT);
		    do_print_menu = 0;
                } else {
                    if (fclose( fp ) != 0)
		    {
			sprintf( err_string, "Could not close '%s'.", current_mat_file );
			error( err_string, "", NOKILL );
		    }
		    remove_trailing_spacers( item_head, pat_string );

        	    if (*filter_exons)
            	  	*num_to_mask = get_exon_info( beg, end, dir, exon_seq, nrow, 
						      exon_file );

		    /* free memory */
                    ckfree( long_string );
		    ckfree( buf );
                    return;
                }
                break;

	    case '1' :		/* List available transcription factors */

		scan_matrix_file( fp, current_mat_file, file_type );
    		fprintf(SCREEN, CHOICE_PROMPT);
		do_print_menu = 0;
		break;

	    case '2' : 		/* Change matrix database */

                fprintf(SCREEN, "\n\nEnter file name for new matrix database\n");
		get_entry (&current_mat_file);  
		change_database( current_mat_file, &file_type, &fp, NOKILL );
                break;

            case '3': 		/* Change minimum number of seqs that must match */

		fprintf(SCREEN, "\n\nEnter new minimum number.   ");
		get_entry (&long_string);
		*min_matches = atoi( long_string );
		while (*min_matches <= 0 || *min_matches > nrow)
		{
		    sprintf( err_string, "Number of sequences that match must be positive"
					 " and less than\n            the number of "
					 "sequences in the alignment (%d).", nrow );
		    error( err_string, "", NOKILL );
                    fprintf(SCREEN, "Enter new minimum number.   ");
		    get_entry (&long_string);
                    *min_matches = atoi( long_string );
		}
                break;

	    case '4' :		/* Change cutoff value */

                fprintf(SCREEN, "\n\nEnter new cutoff value.  ");
		get_entry (&long_string);
                cutoff = atof( long_string );
                while (cutoff <= 0 || cutoff > 1)
                { 
		    error( "Cutoff value must be positive and less than 1.", "", NOKILL );
                    fprintf(SCREEN, "\nEnter new cutoff value.  ");
		    get_entry (&long_string);
                    cutoff = atof( long_string );
		}
		break;

	    case '5' :		/* Specify range */

		get_range( from, to, ncol, highest_pos );
		break;

	    case '6' :		/* Specify which sequences to look in */

		for (i = 0; i < nrow; i++)
                    look_in_seq[i] = 0;
		*min_matches = 0;

		fprintf(SCREEN, "\nEnter the number of each sequence to search in ");
		fprintf(SCREEN, "(in the range 0 to %d),\none sequence per ", nrow - 1);
		fprintf(SCREEN, "line.  After all sequences have been entered, enter\n");
		fprintf(SCREEN, "a '.'  To search on all sequences in the alignment, ");
		fprintf(SCREEN, "initially enter a '.'\n\n");
		get_entry (&long_string);
                i = atoi( long_string );

		while (long_string[0] != '.')
		{
		    if ((i < 0) || (i >= nrow)) {
                  	sprintf( err_string, "Value entered must be between 0 and %d.",
                            (nrow - 1) );
                    	error( err_string, "", NOKILL );
                    	fprintf(SCREEN, "Please enter a new value.  ");
		    } else {
			/* only count sequences that are new (dont count them twice) */
			if (look_in_seq[i] == 0)
			    (*min_matches)++;
                        look_in_seq[i] = 1;
		    }

		    get_entry (&long_string);
                    i = atoi( long_string );
		}

		/* if min_matches is still 0, then the user initially entered a '.' */
		if (*min_matches == 0)
		{
                    for (i = 0; i < nrow; i++)
                        look_in_seq[i] = 1;
                    *min_matches = nrow;
		}
		break;

            case '7' :          /* Toggle exons option */

                *filter_exons = 1 - *filter_exons;
                break;

	    case '8' :		/* Toggle which seq positions are reported in terms of */

		*report_to_top = 1 - *report_to_top;
		break;

	    case '9' :		/* Toggle between masking lowercase nucleotides */

		*count_lc = 1 - *count_lc;
		break;

	    case '0' :		/* Toggle between output formats */

		*out_format = 1 - *out_format;
		break;

	    default :		/* Invalid menu choice */

		print_invalid_menu_choice( choice );
		do_print_menu = 0;
		break;
	}
    }
}


/* Add a spacer to the linked list, given its value(s), add spacer to pattern string */
static void add_spacer( int first, int second, char **pat_string, char *long_string,
			struct ITEM **item_head, int *max_pat_len )
{
    char *p = NULL;		/* char pointer */
    struct ITEM *step, 		/* pointer to step through list of items */
	        *new_item;	/* new item to be added to the list */

    new_item = ckalloc( sizeof( struct ITEM ) );

    /* remove excess chars after spacer if there are any */
    if (!(first == 0 && second == 0))
    {
        for (p = long_string; *p != '>'; p++);
        p++;
        *p = '\0';
    	strcat( *pat_string, long_string );
    	strcat( *pat_string, " " );
    }

    /* fill in info */
    new_item->length = -1;
    new_item->compl = -1;
    new_item->first = first;
    new_item->second = second;
    new_item->max_score = 0;
    new_item->threshold = 0;
    new_item->submat = NULL;
    new_item->next = NULL;

    /* uppdate maximum possible match pattern length */
    *max_pat_len += second;

    /* First element in list */
    if (*item_head == NULL)
        *item_head = new_item;
    else
    {
        step = *item_head;
        while (step->next != NULL)
            step = step->next;
        step->next = new_item;
     }
}


/* Determine if the spacer entered by the user is in the proper format. */
static int can_parse_spacer( char **spacer, int *first, int *second )
{
    char c1, c2, c3,	/* chars for checking format of spacer */
	 *temp_spacer;  /* temporary spacer for reformatting */
    int i;		/* array index */

    /* reformat spacer from '0,100' to '<0,100>' if necessary */
    if (!strstr( *spacer, "<" ) && !strstr( *spacer, ">" ))
    {
    	temp_spacer = ckalloc( sizeof( *spacer ) + 3 * sizeof( char ));
	temp_spacer[0] = '<';
	for (i = 0; (*spacer)[i] != '\0'; i++)
	    temp_spacer[i + 1] = (*spacer)[i];
	temp_spacer[i + 1] = '>';
	temp_spacer[i + 2] = '\0';
	strcpy( *spacer, temp_spacer );
    	free( temp_spacer );
    }

    /* spacer of format <10,20> */
    if (sscanf( *spacer, "%1[<]%d%1[,]%d%1[>]", &c1, first, &c2, second, &c3 ) == 5)
    {
	if ((*first < 0) || (*second < 0) || (*second < *first))
            return 0;
	else
	    return 1;
    }

    /* spacer of format <10> */
    *second = *first;
    if (sscanf( *spacer, "%1[<]%d%1[>]", &c1, first, &c2 ) == 3 )
    {
	if (*first < 0)
	    return 0;
	else
            return 1;
    }

    /* else, invalid spacer */
    return 0;
}


/* Change matrix database to database entered by user. */
static void change_database( char *current_mat_file, char *file_type, FILE **fp,
			     int kill_type )
{
    char *buf;	/* buffer for reading in lines from file to guess its format */

    *file_type = '?';
    buf = ckalloc( BUFSIZE );

    if (fclose( *fp ) != 0)
        error( "Could not close old matrix file.", "", KILL );

    /* while file does not exist or file format is invalid */
    while ((*fp == NULL) || (*file_type == '?'))
    {
    	if (kill_type == KILL)
	    *fp = ckopen( current_mat_file, "r" );
    	else
    	{
    	    *fp = ckopen_nk( current_mat_file, "r" );

	    /* file does not exist, prompt for new filename */
	    if (*fp == NULL)
	    {
        	error( "Invalid file name.", "", NOKILL );
		fprintf(SCREEN, "\nPlease enter a valid file name.  ");
		get_entry (&current_mat_file);
	    }
    	}

	/* file exists, check if it is of proper format */
    	if (*fp != NULL)
    	{
    	    fgets( buf, BUFSIZE, *fp );
    	    *file_type = guess_matrix_type( buf );

    	    if (*file_type == '?')
    	    {
		if (kill_type == KILL)
    	    	    error( "Unknown matrix file type.", "", KILL );

		/* file is of improper format, prompt for new filename */
		else
		{
        	    error( "Invalid file format.", "", NOKILL );
		    fprintf(SCREEN, "\nPlease enter a valid file name.  ");
		    get_entry (&current_mat_file);
		}
    	    }
	    rewind( *fp );
    	}
    }
    ckfree( buf );
}


/* User specifies range to search on the alignment */
static void get_range( int *from, int *to, int ncol, int highest_pos )
{
    char *long_string;	/* utility string to read in user input */

    long_string = ckalloc( STR_SIZE );

    fprintf(SCREEN, "\n\nEnter the beginning of the range.");
    fprintf(SCREEN, "\nTo search throughout the entire alignment, enter a -1.  ");
    get_entry (&long_string);
    *from = atoi( long_string );
    while ((*from == 0) || (*from < -1) || (*from > highest_pos))
    {
	sprintf( err_string, "Position must be in the range 1 to %d.", highest_pos );
	error( err_string, "", NOKILL );
        fprintf(SCREEN, "Enter new beginning of the range.");
        fprintf(SCREEN, "\nTo search throughout the entire alignment, enter a -1.  ");
	get_entry (&long_string);
        *from = atoi( long_string );
    }

    /* search over entire range. */
    if (*from == -1)
    {
	*from = 1;
	*to = ncol;
    } else {
    	fprintf(SCREEN, "\nEnter the end of the range.\nEnter a '.' to search");
    	fprintf(SCREEN, " to the end of the alignment.  ");
	get_entry (&long_string);
   	if (long_string[0] == '.')
            *to = ncol;
    	else
    	{
            *to = atoi( long_string );
            while (*to <= 0 || *to > highest_pos || *to < *from)
            {
            	sprintf( err_string, "Position must be in range 1 to %d", highest_pos );
	    	strcat( err_string, "\n          and greater than the start position.");
            	error( err_string, "", NOKILL );
            	fprintf(SCREEN, "Enter new end position.  ");
		get_entry (&long_string);
            	if (long_string[0] == '.')
		{
            	    *to = ncol;
		    break;
		}
            	else
                    *to = atoi( long_string );
            }
    	}
    }
    ckfree( long_string );
}


/* Scan through matrix file and write available names/patterns to file. */
static void scan_matrix_file( FILE *fp, char *current_mat_file, char file_type )
{
    char chr, 			/* single char read in from buffer */
	 *buf,			/* buffer read in from one line of the matrix file */
	 *name,			/* name, */
	 *pat,			/* pattern, */ 
	 *acnum,		/* and accession number of factor, as read in from file */
	 *file_name;		/* name of file to output factors to */
    float num1, num2;		/* values read in from the buffer */
    FILE *listfp = NULL;	/* file pointer for output file */

    /* allocate memory */
    buf = ckalloc( BUFSIZE );
    name = ckalloc( STR_SIZE );
    pat = ckalloc( STR_SIZE );
    acnum = ckalloc( STR_SIZE );
    file_name = ckalloc( STR_SIZE );

    strcpy( file_name, current_mat_file );
    strcat( file_name, "_list" ); 
    rewind( fp );

    /* open file and print header */
    listfp = ckopen( file_name, "w" );
    fprintf(listfp, "Available transcription factors in file '%s':\n\n\n", 
		     current_mat_file);
    if (file_type == 'i')
    {
    	fprintf(listfp, "%-10s%-20s%s\n", "AC NUM", "NAME", "CONSENSUS PATTERN");
    	fprintf(listfp, "%-10s%-20s%s\n\n", "------", "----", "-----------------");
    } else {
	fprintf(listfp, "%-10s%-20s%-20s%s\n","AC NUM","ID","NAME","CONSENSUS PATTERN");
        fprintf(listfp, "%-10s%-20s%-20s%s\n\n","------","--","----","-----------------");
    }

    switch (file_type)
    {
	/* imd database */
	case 'i' :

	    while (fgets( buf, BUFSIZE, fp ))
    	    {
        	if (sscanf( buf, "%s %f %f %s %s", name, &num1, &num2, pat, acnum ) == 5)
		{
  	    	    fprintf(listfp, "%-10s%-20s%s\n", acnum, name, pat);
		    while (fgets( buf, BUFSIZE, fp ))	/* skip nucleotide values */
		    	if (starts( skip_ws( buf ), "T |" ))
			    break;
		}
    	    }
	    break;
	
	/* transfac database */
	case 't' :

            while (fgets( buf, BUFSIZE, fp ))
	    {
		if (sscanf( buf, "NA %s", name ) == 1)
		    fprintf(listfp, "%-20s", name);
                else if (sscanf( buf, "AC %s", acnum ) == 1)
                    fprintf(listfp, "%-10s", acnum);
                else if (sscanf( buf, "ID %s", acnum ) == 1)
                    fprintf(listfp, "%-20s", acnum);
		else if (sscanf( buf, "P0 %s", name ) == 1)
		{
                    pat[0] = '\0';
            	    while ((fgets( buf, BUFSIZE, fp )) && (!strstr( buf, "XX" ))) 
		    {
			sscanf( buf, "%f %f %f %f %f %s", 
				&num1, &num1, &num1, &num1, &num1, &chr );
			strcat( pat, &chr );
		    }
		    fprintf(listfp, "%s\n", pat);
		}	
	    }
	    break;

	default :

	    error( "1 Internal 'file_type' error.", "", KILL );
    }

    if (fclose( listfp ) != 0)
    {
	sprintf( err_string, "Could not close '%s'.\n", file_name ); 
	error( err_string, "", NOKILL );
    }

    fprintf(SCREEN, "\nThe available motifs have been written to ");
    fprintf(SCREEN, "file '%s'\n\n", file_name);

    /* rewind file pointer and free memory */
    rewind( fp );
    ckfree( buf );
    ckfree( name );
    ckfree( pat );
    ckfree( acnum );
    ckfree( file_name );
}


/* Find where the relevant information is in the matrix file. */
static int get_data_from_file( FILE *fp, char *string, int type, struct ITEM **item_head,
			       char **pat_string, float cutoff, char file_type,
			       int *num_motifs, int *max_pat_len )
{
    char chr,                   /* single char read in from buffer */
         *buf,	          	/* buffer read in from one line of the matrix file */
         *name,                 /* name, */
         *pat,                  /* pattern, */
         *acnum,                /* and accession number of factor, as read in from file */
	 *p = NULL;		/* string pointer for processing reverse complements */
    int found = 0, 		/* was the search query found in the file? (bool) */
	compl = 0, 		/* is the motif on the reverse strand? (bool) */
	num_int;		/* int read in from buffer */
    float num_float1, 		/* floats read in from buffer */
	  num_float2;
    long pos_in_file;		/* position to return back to in matrix file */

    /* allocate memory */
    buf = ckalloc( BUFSIZE );
    name = ckalloc( STR_SIZE );
    pat = ckalloc( STR_SIZE );
    acnum = ckalloc( STR_SIZE );

    /* process complements */
    for (p = string; *p != '\0'; p++)
        if (*p == '(' && *(p + 1) == 'c' && *(p + 2) == ')')
        {
	    compl = 1;
	    *p = '\0';
	}

    switch (file_type)
    {
	/* imd database */
	case 'i' :

    	    while (fgets( buf, BUFSIZE, fp ))
    	    {
        	if (sscanf( buf, "%s %f %f %s %s", name, &num_float1, &num_float2, 
						   pat, acnum ) == 5)
		{
		    if (((type == NAME) && (strcasecmp( string, name ) == 0)) || 
	    	    	((type == PAT) && (strcasecmp( string, pat ) == 0)) ||
                    	((type == AC) && (strcasecmp( string, acnum ) == 0)))
		    {
		    	add_motif_to_list( compl, string, pat_string, cutoff, fp, 
				           file_type, item_head, max_pat_len );
	    	        found = 1;
	    	    	break;
		    }
		}
    	    }
	    break;

	/* transfac database */
	case 't' :

	    if ((type == NAME) || (type == AC) || (type == ID))
	    {
            	while (fgets( buf, BUFSIZE, fp ))
	    	{
                    if (((type == NAME) && (sscanf( buf, "NA %s", name ) == 1) && 
		         (strcasecmp( string, name ) == 0)) ||
			((type == AC) && (sscanf( buf, "AC %s", acnum ) == 1) &&
                         (strcasecmp( string, acnum ) == 0)) ||
			((type == ID) && (sscanf( buf, "ID %s", acnum ) == 1) &&
                         (strcasecmp( string, acnum ) == 0)))
		    {
                    	while ((fgets( buf, BUFSIZE, fp )) && (!strstr( buf, "P0" ))) { }
		    
                    	add_motif_to_list( compl, string, pat_string, cutoff, fp, 
                                           file_type, item_head, max_pat_len );
                        found = 1;
                        break;
		    }
	        }
	    } 
	    else if (type == PAT)
	    {
		while (fgets( buf, BUFSIZE, fp ))
		{
                    if ((sscanf( buf, "P0 %s", name ) == 1))
                    {
                    	pat[0] = '\0';

			/* record position in the file */
			pos_in_file = ftell( fp );
                        while ((fgets( buf, BUFSIZE, fp )) && (!strstr( buf, "XX" )))
                    	{
                            if (sscanf( buf, "%d %f %f %f %f %s",
                                    &num_int, &num_float1, &num_float1, &num_float1, 
				    &num_float1, &chr ) != 6)
				error( "Improper Transfac format.", buf, KILL );	
                            strcat( pat, &chr );
                        }
		        if (strcasecmp( string, pat ) == 0)
		        {
			    /* return to previously recorded place in the file */
			    fseek( fp, pos_in_file, SEEK_SET );
                    	    add_motif_to_list( compl, string, pat_string, cutoff, fp,
                                               file_type, item_head, max_pat_len );
			    found = 1;
			    break;
		        }
		    }
		}
	    }
	    break;

	default :

            error( "2 Internal 'file_type' error.", "", KILL );
    }

    rewind( fp );

    /* free allocated memory */
    ckfree( buf );
    ckfree( name );
    ckfree( pat );
    ckfree( acnum );

    if (!found)
    {
        error( "Search query not found in file.", "", NOKILL );
	return 0;
    } else {
	(*num_motifs)++;
    	return 1;
    }
}


/* Fill in information into the item structure (in this case a motif) and add it
   to the linked list. */
static void add_motif_to_list( int compl, char *string, char **pat_string, float cutoff,
			       FILE *fp, char file_type, struct ITEM **item_head,
			       int *max_pat_len )
{
    struct ITEM *step = NULL, 	/* pointer for stepping through item list */
	        *new_item;	/* new item to be added to the list */

    new_item = ckalloc( sizeof( struct ITEM ) );

    /* add motif to pattern */
    if (compl)
        strcat( string, "(c)" );
    sprintf( *pat_string, "%s{%s[%.2f]: %s} ", *pat_string, string, cutoff, 
	     file_type == 't' ? "TransFac" : "IMD" );

    /* fill in info */
    new_item->compl = compl;
    new_item->first = 0;
    new_item->second = 0;
    new_item->next = NULL;
    read_matrix( fp, new_item, file_type );
    new_item->max_score = get_max_score( new_item );
    new_item->min_score = get_min_score( new_item );

    *max_pat_len += new_item->length;

    /* equation taken from the MOTIF program (http://motif.genome.ad.jp) */
    new_item->threshold = cutoff * new_item->max_score + 
			  (1 - cutoff) * new_item->min_score;

    /* add item to list */
    if (*item_head == NULL)        	/* First element in list */
        *item_head = new_item;
    else			   	/* List already exists */
    {
        step = *item_head;
        while (step->next != NULL)
            step = step->next;
        step->next = new_item;
    }
}


/* Given the first line of the matrix file, guess what format it is.  Returns 'i' for
   for the imd database, 't' for transfac, 'c' for consensus, and '?' for unknown. */
static char guess_matrix_type( char *buf )
{
    char type = '?',		/* guessed matrix type */
	 *name,                 /* name, */
         *pat,                  /* pattern, */
         *acnum;                /* and accession number of factor, as read in from file */
    float num_float1,           /* floats read in from buffer */
	  num_float2;

    /* allocate memory */
    name = ckalloc( STR_SIZE );
    pat = ckalloc( STR_SIZE );
    acnum = ckalloc( STR_SIZE );

    /* determine type */
    if (sscanf( buf, "%s %f %f %s %s", name, &num_float1, &num_float2, pat, acnum ) == 5)
	type = 'i';
    else if ((sscanf( buf, "AC %s", acnum ) == 1) || 
	     (starts( skip_ws( buf ), "VV  TRANSFAC" )))
	type = 't';
    else if ((sscanf( buf, "%s %s", name, pat ) == 2))
	type = 'c';

    /* free allocated memory */
    free( name );
    free( pat );
    free( acnum );

    return type;
}


/* Given a file pointer to the beginning of the matrix, read matrix values from the
   matrix file directly into the given structure.  Also record 'length' field */
static void read_matrix( FILE *fp, struct ITEM *new_item, char file_type )
{
    char chr,                   /* single char read in from buffer */
         *buf,         		/* buffer read in from one line of the matrix file */
	 *p;			/* pointer for moving through buffer */
    int row, 			/* current row of matrix (rowA, rowC, rowG, rowT) */
	pos, 			/* current position in the motif */
	val = 0, 		/* freq at that position in motif for that nucleotide */
	adjust = 0, 		/* total of each nucleotide freq at one position */
	reached_end = 0;	/* have we reached the end of the line? (bool) */
    float fA, fC, fG, fT;	/* freq for each nucleotide at one position */

    buf = ckalloc( BUFSIZE );
    (*new_item).submat = (float **) ckalloc( sizeof( float*) * 4 );
    for (pos = 0; pos < 4; pos++)
	(*new_item).submat[pos] = (float *) ckalloc( sizeof( float ) * MAX_SUBMAT_LEN );

    switch (file_type)
    {
	/* imd database */
	case 'i' :

    	    for (row = rowA; row <= rowT; row++) 
    	    {
		fgets( buf, BUFSIZE, fp );
		p = skip_ws( buf );
		if (!strchr( "ACGT", *p ))
		    error( "Invalid matrix file format.", "", KILL );
		p += 3; 	/* skip 'A |' */
		pos = 0;
		reached_end = 0;

		/* process each int value */
		while (!reached_end)
		{
		    if (pos > MAX_SUBMAT_LEN)
			error( "Motif in matrix file too long to process.", "", KILL ); 

		    /* skip spaces, account for trailing spaces */
		    while (*p == ' ')
		    {
		        p++;
			if (*p == '\n')
			    reached_end = 1;
		    }
		    if (!reached_end)
		    {
		    	val = atoi( p );
		    	pos++;
                    	new_item->submat[row][pos] = val;

			/* skip remainder of number, if it is more than 1 digit */
                    	while (*p != '\n' && *p != ' ')
                            p++;
                    	if (*p == '\n')
                            reached_end = 1;
		    }
		}
		adjust += val;
    	    }
	    break;

	/* transfac database */
	case 't' :

            while ((fgets( buf, BUFSIZE, fp )) && (buf[0] != 'X'))
	    {
		if ((sscanf(buf,"%d %f %f %f %f %c", &pos, &fA, &fC, &fG, &fT, &chr)) !=6)
		    error( "Invalid transfac matrix file format.", buf, KILL );
                if (pos > MAX_SUBMAT_LEN)
		    error( "Motif in matrix file too long to process.", "", KILL );

		new_item->submat[rowA][pos] = fA;
                new_item->submat[rowC][pos] = fC;
                new_item->submat[rowG][pos] = fG;
                new_item->submat[rowT][pos] = fT;
	    }
            adjust = fA + fC + fG + fT;
	    break;

	default :

            error( "3 Internal 'file_type' error.", "", KILL );
    }

    new_item->length = pos;
    ckfree( buf );
}
	

/* Compute the maximum possible score that a certain motif can achieve. */ 
static float get_max_score( struct ITEM *item )
{
    int row, pos;       	/* current row and position in matrix */
    float max, 			/* maximum value at one position */
	  tot_max = 0;		/* total of all maximum values across all positions */

    for (pos = 1; pos <= item->length; pos++)
    {
	max = INT_MIN;
	for (row = rowA; row <= rowT; row++)
	    if (item->submat[row][pos] > max)
		max = item->submat[row][pos];
	tot_max += max;
    }

    return (tot_max);
}


/* Compute the minimum possible score that a certain motif can achieve. */
static float get_min_score( struct ITEM *item )
{
    int row, pos;               /* current row and position in matrix */
    float min,                  /* minimum value at one position */
          tot_min = 0;          /* total of all minimum values across all positions */

    for (pos = 1; pos <= item->length; pos++)
    {
        min = INT_MAX;
        for (row = rowA; row <= rowT; row++)
            if (item->submat[row][pos] < min)
                min = item->submat[row][pos];
        tot_min += min;
    }

    return (tot_min);
}


/* Remove the last node from the linked list. */
static void delete_last_node( struct ITEM **item_head, int *num_motifs )
{
    int i;			/* array index */
    struct ITEM *step;		/* pointer to step through item list */

    step = *item_head;
    if (!step)
	error( "Internal error: empty ITEM list.", "", KILL );

    if (step->next != NULL)
    {
	/* move to end of list */
    	while (step->next->next != NULL)
            step = step->next;

	/* if its a motif, decrement num_motifs, and free submat */
    	if (step->next->length != -1)
	{
            (*num_motifs)--;
	    for (i = 0; i < 4; ++i)
	        ckfree( step->next->submat[i] );
	    ckfree( step->next->submat );
	}

    	ckfree( step->next );
    	step->next = NULL;
    }
}


/* Remove the last pattern from the pattern string. */
static void delete_last_pat( char **pat_string )
{
    char *end, *start;		/* pointers for moving through the pattern string */

    start = end = *pat_string;
    while (1)
    {
	while ((*end != '}' && *end != '>'))
   	    end++;
	end += 2;
	if ((*end == '{') || (*end == '<'))
            start = end;
	else
	{
	    *start = '\0';
	    break;
	}
    }
}


/* Remove all trailing spacers from pattern (spacers occurring after the final motif.) */
static void remove_trailing_spacers( struct ITEM **item_head, char **pat_string )
{
    char *char_prev_motif = NULL, 	/* pointer to previous motif string */
	 *char_item_step;		/* pointer to step through pattern string */
    struct ITEM *prev_motif = NULL, 	/* pointer to previous motif structure */
		*item_step;		/* pointer to step through item list */

   char_item_step = *pat_string;
   item_step = *item_head;

    /* update linked list */
    while (item_step != NULL)
    {
	if (item_step->length != -1)	/* save location of previous motif */
	    prev_motif = item_step;
	item_step = item_step->next;
    }
    free_item_list( &prev_motif->next );
    prev_motif->next = NULL;

    /* update pattern */
    while (*char_item_step)
    {
	if (*char_item_step == '}')
	    char_prev_motif = char_item_step;
	char_item_step++;
    }
    char_prev_motif += 2;
    if (*char_prev_motif == '<')	/* if we had to remove a trailing spacer */
	error( "Trailing spacers in pattern automatically removed.", "", WARN);
    *char_prev_motif = '\0';
}


/* Print summary at the top of the file. */
static void print_header( char *align_file, int from, int to, char *pat_string, int nrow,
			  int ncol, int filter_exons, int exon_seq, int *look_in_seq,
			  int min_matches, char align_type, int report_to_top, 
			  int num_motifs, char *exon_file, int count_lc, int out_format )
{
    int i, 		/* array index */
	num_seqs = 0;	/* number of seqs to look in */

    if (out_format == GFF_OUT)
	printf ("# ");
    printf("Hits in alignment file '%s' within the range of ", align_file);
    if (to == ncol)
	if (from == 1)
	    printf("the entire alignment\n");
	else 
	    printf("%d to the end\n", from);	
    else
	printf("%d to %d\n", from, to);

    if (out_format == GFF_OUT)
        printf ("# ");
    printf("in relation to %s", report_to_top ? "the top-most sequence\n" : 
						"the sequence that the hit occurs on\n");
    if (out_format == GFF_OUT)
        printf ("# ");
    printf("for the pattern \"%s\",\n", pat_string);
    if (filter_exons)
    {
    	if (out_format == GFF_OUT)
            printf ("# ");
	printf("filtering out exon regions provided in the file '%s' ", exon_file);
	printf("(in relation to sequence %d),\n", exon_seq);
    }

    /* count how many sequences will be searched in to format output */
    for (i = 0; i < nrow; i++)
	if (look_in_seq[i])
	    num_seqs++;

    /* tell how lowercase nucs are treated */
    if (out_format == GFF_OUT)
        printf ("# ");
    if (count_lc)
        printf("treating lowercase nucleotides the same as uppercase nucleotides,\n");
    else
        printf("treating lowercase nucleotides as being masked out,\n");

    if (out_format == GFF_OUT)
        printf ("# ");
    if (num_seqs == 1) {
        printf("where at least 1 of the %d sequences matches in sequence", nrow);
        for (i = 0; i < nrow; i++)
       	    if (look_in_seq[i])
            	printf(" %d.", i);
    } else {
        printf("where at least %d of the %d sequences match%s in sequences", 
               min_matches, nrow, (min_matches == 1 ? "es" : ""));
        for (i = 0; i < nrow; i++)
    	{
            if (look_in_seq[i])
	    {
	    	if (num_seqs == 1)
                    printf(" and %d:", i);
	  	else if (num_seqs == 2)
		    printf(" %d", i);
	        else
		    printf(" %d,", i);

		num_seqs--;
	    }
	}
    }

    printf("\n\n\n");
    if (out_format == TFFIND_OUT)
    {
    if (align_type == 'b')
    {
    	printf("%-4s %-8s %-8s%-7s", "Seq", "From", "To", "Blocks");
	for (i = 1; i < num_motifs; i++)
	    printf("%-7s", " ");
	printf(" %-4s  %-4s  %s\n", "Strand", "Score", "Pattern");
        printf("%-4s %-8s %-8s%-7s", "---", "----", "--", "------");
        for (i = 1; i < num_motifs; i++)
            printf("%-7s", " ");
	printf(" %-4s  %-4s  %s\n\n", "------", "-----", "-------");
    } else {
        printf("%-4s %-8s %-8s %-7s %-4s  %s\n", "Seq", "From", "To", "Strand", 
               "Score", "Pattern");
        printf("%-4s %-8s %-8s %-7s %-4s  %s\n\n", "---", "----", "--", "------",
               "-----", "-------");
    }
    }
}


/* Search for pattern in the current dir, in the given range, on the proper sequences. */
static void search_for_pattern( char **A, int ncol, int nrow, struct ITEM *item_head, 
				int num_motifs, int from, int to, int cur_dir,
				int *look_in_seq, struct MATCH **match_head,
				char align_type, Block_ptr block_head, Contig_ptr contig,
				int **col2pos, int report_to_top, int min_matches,
				int count_lc )
{
    int i, j;				/* array indices */
    Block_ptr b_ptr;			/* pointer to step through block list */

    /* multipipmaker and FASTA format */
    if ((align_type == 'm' ) || (align_type == 'f'))
    {
	/* reverse sequences in place for each */
    	if (cur_dir == REVERSE)
            for (i = 0; i < nrow; i++)
            	do_revcomp( A[i], ncol );

	/* for each sequence in the alignment, record matches at each position */
    	for (i = 0; i < nrow; ++i)
	    if (look_in_seq[i])			/* if we should search on this sequence */
            	for (j = from; j <= to; ++j)
		    record_pattern_matches( match_head, item_head, A[i], nrow, ncol, i,
					    j, num_motifs, cur_dir, align_type, NULL, 
					    NULL, NULL, col2pos, report_to_top, count_lc);
    /* blockbuilder format */
    } else {
        /* reverse sequences in place for each */
        if (cur_dir == REVERSE)
	    for (b_ptr = block_head; b_ptr; b_ptr = b_ptr->next)
                for (i = 0; i < b_ptr->nrow; i++)
                    do_revcomp( b_ptr->seg[i].chars, b_ptr->ncol );

	/* for each sequence in the block */
    	for (b_ptr = block_head; b_ptr; b_ptr = b_ptr->next)
        {
	    /* don't bother looking in this block if there is no chance for a match */
	    if (b_ptr->nrow < min_matches)
		continue;

            for (i = 0; i < b_ptr->nrow; ++i)
	    {
		/* don't bother looking if we are outside of the range */
		if ((b_ptr->seg[0].end < from) || (b_ptr->seg[0].beg > to))
		     break;

		if (look_in_seq[b_ptr->seg[i].contig_nbr])
		{
		    /* for each column in the alignment */
                    for (j = 0; j < b_ptr->ncol; ++j)
                    	record_pattern_matches( match_head, item_head, NULL, nrow,
                                                b_ptr->ncol, i, j, num_motifs, cur_dir,
						align_type, b_ptr, block_head, contig,
						col2pos, report_to_top, count_lc );
		}
	    }
        }
    }
}


/* Record all matches to the pattern at the given sequence and position. */
static void record_pattern_matches( struct MATCH **match_head, struct ITEM *item_head,
				    char *seq, int nrow, int ncol, int which_row,
				    int start_pos, int num_motifs_in_pat, int cur_dir,
				    char align_type, Block_ptr b_ptr,
                                    Block_ptr block_head, Contig_ptr contig,
				    int **col2pos, int report_to_top, int count_lc )
{
    int m, 			/* array index */
	which_match = 0, 	/* which match we are currently on to the pattern */
	which_motif = 0, 	/* which motif we are currently on in the pattern */
	startpt, endpt, 	/* the endpoints of the match for output */
	match_len,		/* length of the entire match, in bps */
	offset = 0, 		/* where we should start looking for the next motif */
	tot_matches, 		/* total matches found (excluding current iteration) */
	found_match = 1,	/* did we find any matches to the current motif? (bool) */
	gap_in_top, 		/* does the hit align to a gap in the top sequence? */
	start, end,		/* window to look in for the next motif */
        prev_motif_len = 0,     /* length of the previous motif in the pattern */
	**match_arr,		/* array giving position of each motif in the alignment */
	**bp_arr,		/* array giving position of each motif rel. to block */
	**block_nbr_arr;	/* array giving block numbers each match runs through */
    float **pcts_arr;		/* array giving pct of each motif in a match */
    struct ITEM **item_step;	/* pointer to step through item list */

    item_step = &item_head;

    /* allocate memory */
    match_arr = (int **) ckalloc( sizeof( int *) * 1 );
    bp_arr = (int **) ckalloc( sizeof( int *) * 1 );
    block_nbr_arr = (int **) ckalloc( sizeof( int *) * 1 );
    pcts_arr = (float **) ckalloc( sizeof( float *) * 1 );
    match_arr[0] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
    bp_arr[0] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
    block_nbr_arr[0] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
    pcts_arr[0] = (float *) ckalloc( sizeof( float ) * num_motifs_in_pat );

    /* initialize arrays */
    for (m = 0; m < num_motifs_in_pat; m++)
    {
	match_arr[0][m] = -1;
	bp_arr[0][m] = -1;
	block_nbr_arr[0][m] = -1;
	pcts_arr[0][m] = -1;
    }

    /* look for 1st match, and iteratively search within window for subsequent matches */
    while ((found_match) && (get_next_window( &start, &end, item_step )))
    {
	found_match = 0;
	which_match = 0;
	tot_matches = which_match + 1;

	for (m = 0; m < tot_matches; m++)
	{
	    /* assign a value to offset based on the file type, which motif we are on in
	       the pattern, and the strand */
	    if (b_ptr)

		/* its not the first motif in the pattern */
		if (which_motif != 0)
		    if (cur_dir == FORWARD)
			offset = match_arr[m][which_motif - 1] + (prev_motif_len + 1);
		    else
			offset = match_arr[m][which_motif - 1] - (prev_motif_len + 1);

		/* it is the first motif in the pattern */
		else
		    offset = b_ptr->seg[which_row].beg - 1 + start_pos;

	    /* type 'm' and 'f' files */
	    else
		/* its not the first motif in the pattern */
		if (which_motif != 0)
		    offset = match_arr[m][which_motif - 1] + (prev_motif_len + 1);

		/* it is the first motif in the pattern */
	  	else
		    offset = start_pos;

	    /* record all matches to the current motif in the current window */
	    if (record_matches_in_window( start, end, offset, m, item_step, seq, ncol,
			   	          &match_arr, &which_match, which_motif,
				  	  num_motifs_in_pat, &pcts_arr, b_ptr,
					  &block_nbr_arr, block_head, contig, which_row,
					  cur_dir, &bp_arr, prev_motif_len, count_lc ))
		found_match = 1;
	}
	which_motif++;
        prev_motif_len = (*item_step)->length - 1;
    }
    if (found_match)
    {
    	get_lowest_pcts( pcts_arr, which_match + 1, num_motifs_in_pat);

        /* debug */ /*
	int n;
    	for (m = 0; m < which_match + 1; m++)
            for (n = 0; n < num_motifs_in_pat; n++)
	    {
            	printf("MATCH ARR [%d][%d] is %d\n", m, n, match_arr[m][n]);
            	printf("BP ARR [%d][%d] is %d\n", m, n, bp_arr[m][n]);
	    }
	*/

    	for (m = 0; m < which_match + 1; m++)
    	{
	    if (b_ptr && cur_dir == REVERSE)
		match_len = (match_arr[m][0] + (*item_step)->length -
		    	     match_arr[m][which_motif - 1]);	
	    else
    	    	match_len = (match_arr[m][which_motif - 1] + (*item_step)->length - 
		             match_arr[m][0]);

	    /* only count as a match if all motifs in the pattern are present */
	    if (valid_match( match_arr[m], num_motifs_in_pat ))
	    {
	    	if ((align_type == 'm') || (align_type == 'f'))
	 	    get_endpoints( match_arr[m][0], cur_dir, match_len, ncol, 
			           col2pos[(report_to_top ? 0 : which_row)], &startpt,
			           &endpt, &gap_in_top, align_type, NULL, NULL, NULL,
			           which_row, report_to_top );
	    	else
            	    get_endpoints( match_arr[m][0], cur_dir, match_len, ncol, NULL,
                                   &startpt, &endpt, &gap_in_top, align_type, b_ptr,
			           block_head, contig, which_row, report_to_top );

		/* add match to match list */
            	add_to_matches( num_motifs_in_pat, 
			        (b_ptr == NULL ? which_row 
					       : b_ptr->seg[which_row].contig_nbr),
		                nrow, match_head, match_len, pcts_arr[m][0], 
			        block_nbr_arr[m], startpt, endpt, gap_in_top, 
				report_to_top, bp_arr[m], cur_dir );
	    }
    	}
    }

    /* free allocated memory */
    free_int_arr( match_arr, which_match + 1 );
    free_int_arr( bp_arr, which_match + 1 );
    free_int_arr( block_nbr_arr, which_match + 1 );
    free_float_arr( pcts_arr, which_match + 1 );
}


/* Get next spacer window to search for the next motif in. Move item_step to next motif.*/
static int get_next_window( int *start, int *end, struct ITEM **item_step )
{
    *start = (*item_step)->first;
    *end = (*item_step)->second;

    while ((*item_step)->next != NULL)
    {
        *item_step = (*item_step)->next;
	if ((*item_step)->length == -1)		/* its a spacer */
	{
	    *start += (*item_step)->first;
	    *end += (*item_step)->second;
	}
	else					/* its a motif */
	    return 1;
    }

    return 0;					/* no motifs remain in the pattern */
}


/* Look in window for all matches to the current motif. */
static int record_matches_in_window( int start, int end, int offset, int cur_match,
			      	     struct ITEM **item_step, char *seq, int ncol,
				     int ***match_arr, int *which_match, int which_motif,
				     int num_motifs_in_pat, float ***pcts_arr,
				     Block_ptr b_ptr, int ***block_nbr_arr, 
				     Block_ptr block_head, Contig_ptr contig, 
				     int which_row, int cur_dir, int ***bp_arr,
				     int prev_motif_len, int count_lc )
{
    int i,			/* used to find block in array */	
	cp,			/* position relative to the entire sequence */
	bp, 			/* position relative to the current block */
    	which_seq = 0,		/* current sequence that we are on */
	bound = -1,		/* boundary for cp for 'b type' files */
	init_start,		/* intialize value for start, before any changes */
	found_a_match = 0;	/* did we find at least one match to the motif? (bool) */
    float score;		/* motif score at the current position */

    init_start = start;
    end += offset;

    /* assign cp and bp based on current case */
    if (b_ptr)
    {
        which_seq = b_ptr->seg[which_row].contig_nbr;
	if (which_motif)
	{
            b_ptr = block_head;
	    for (i = 1; i < (*block_nbr_arr)[*which_match][which_motif - 1]; i++)
                b_ptr = b_ptr->next;
            which_row = get_row_from_seq_num( which_seq, b_ptr->nrow, b_ptr );
	}
        ncol = b_ptr->ncol;

        if (cur_dir == REVERSE)
        {
            if (which_motif)
                cp = offset - start;
            else
                cp = 2 * (b_ptr->seg[which_row].beg - 1) + b_ptr->ncol - 1 -
		     (start + offset);
            bound = cp - (end - (start + offset));
        } else
            cp = (start + offset);

        if (which_motif)
	    bp = (*bp_arr)[*which_match][which_motif - 1] + init_start + 1 + 
		 prev_motif_len;
        else
            bp = (start + offset) - b_ptr->seg[which_row].beg + 1;

    } else {
	cp = (start + offset);
	bp = (start + offset);
    }

    /* search within window for all matches */
    for( ; ((b_ptr && cur_dir == REVERSE) ? cp >= bound: cp <= end);
	   ((b_ptr && cur_dir == REVERSE) ? cp -- : cp++), bp++ )
    {
	if (bp >= ncol) /* ran out of room on this block */
	{
	    if (b_ptr == NULL)
	    	return (found_a_match);
	    else {
		bp -= b_ptr->ncol;
	        move_to_adjacent_block( &b_ptr, &ncol, block_head, contig, &which_row,
			            	cur_dir );
                if (b_ptr == NULL)
                    return (found_a_match);
	    }
	}
	/* printf("CP %d BP %d CHAR %c\n", cp, bp, seq[bp]); */
        score = motif_score( seq, *item_step, bp, ncol, b_ptr, block_head, contig, 
			     which_row, cur_dir, count_lc );

        if (score >= ((*item_step)->threshold))
	{
	    /* debug */ /*
	    printf("\n\nMATCH! at CP %d BP %d OFF %d BEG %d WHICH MO %d WHICH MA %d\n", 
		   cp, bp, offset, b_ptr->seg[which_row].beg, which_motif, *which_match );
	    printf("\n\nMATCH! at CP %d OFF %d WHICH MO %d WHICH MA %d SEQ %d\n", 
		   cp, offset, which_motif, *which_match, which_seq);
	    */
            record_match( match_arr, which_match, which_motif, cp, score, 
                          (*item_step)->max_score, pcts_arr, num_motifs_in_pat, cur_match,
			  block_nbr_arr, (b_ptr ? b_ptr->mark : 0), bp, bp_arr,
			  (*item_step)->min_score );
            found_a_match = 1;
	}
    }

    return (found_a_match);
}


/* Move block pointer and other relavent info to adjacent block to continue search. */
static void move_to_adjacent_block( Block_ptr *block, int *ncol, Block_ptr block_head,
				    Contig_ptr contig, int *which_row, int cur_dir )
{
    int	which_seq; 		/* number of the sequence we are currently dealing with */
    Seg_ptr s;			/* ptr used for calls to find_block and adjacent_block */

    which_seq = (*block)->seg[(*which_row)].contig_nbr;
    find_block( block_head, which_seq, (*block)->seg[*which_row].beg, contig, &s );
    *block = adjacent_block( &s, contig, block_head, cur_dir );

    /* return if we have run off the end or beginning of the alignment */
    if (*block == NULL)
	return;

    *ncol = (*block)->ncol;
    *which_row = get_row_from_seq_num( which_seq, (*block)->nrow, *block );
    if (*which_row == -1)
	error( "Requested seq not found in given block in move_to_adjacent_block.", 
	       "", KILL );
}


/* Record position and pct of one match. */
static void record_match( int ***match_arr, int *which_match, int which_motif,
			  int cur_pos, float score, float max_score, float ***pcts_arr,
			  int num_motifs_in_pat, int cur_match, int ***block_nbr_arr,
			  int block_nbr, int bp, int ***bp_arr, float min_score )
{
    int m, n, 				/* array indices */
	**temp_match_arr = NULL, 	/* temporary arrays for creating new arrays */
	**temp_bp_arr = NULL,
	**temp_block_arr = NULL;
    float lowest_cutoff,
	  **temp_pcts_arr = NULL;

    /* if its not the first motif in the first match */
    if ((*which_match || which_motif) && ((*match_arr)[cur_match][which_motif] != -1))
    {
	/* move to next match */
        (*which_match)++;

	/* allocate temporary arrays */
    	temp_match_arr = (int **) ckalloc( sizeof( int *) * (*which_match) );
        temp_bp_arr = (int **) ckalloc( sizeof( int *) * (*which_match) );
        temp_pcts_arr = (float **) ckalloc( sizeof( float *) * (*which_match) );
 	temp_block_arr = (int **) ckalloc( sizeof( int *) * (*which_match) );
        for (m = 0; m < (*which_match); m++)
   	{
	    temp_match_arr[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
            temp_bp_arr[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
            temp_pcts_arr[m] = (float *) ckalloc( sizeof( float ) * num_motifs_in_pat );
            temp_block_arr[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
	}

	/* copy info into temporary arrays */
    	for (m = 0; m < (*which_match); m++)
	{
            for (n = 0; n < num_motifs_in_pat; n++)
	    {
                temp_match_arr[m][n] = (*match_arr)[m][n];
                temp_bp_arr[m][n] = (*bp_arr)[m][n];
                temp_pcts_arr[m][n] = (*pcts_arr)[m][n];
                temp_block_arr[m][n] = (*block_nbr_arr)[m][n];
	    }
	}

	/* clean up old real arrays */
	free_int_arr( *match_arr, *which_match);
	free_int_arr( *bp_arr, *which_match);
	free_float_arr( *pcts_arr, *which_match);
	free_int_arr( *block_nbr_arr, *which_match);

	/* start reallocating real arrays */
        (*match_arr) = (int **) ckalloc( sizeof( int *) * (*which_match + 1) );
        (*bp_arr) = (int **) ckalloc( sizeof( int *) * (*which_match + 1) );
        (*pcts_arr) = (float **) ckalloc( sizeof( float *) * (*which_match + 1) );
        (*block_nbr_arr) = (int **) ckalloc( sizeof( int *) * (*which_match + 1) );

	/* finish reallocating real arrays */
	for (m = 0; m < (*which_match + 1); m++)
	{
    	    (*match_arr)[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
            (*bp_arr)[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
            (*pcts_arr)[m] = (float *) ckalloc( sizeof( float ) * num_motifs_in_pat );
            (*block_nbr_arr)[m] = (int *) ckalloc( sizeof( int ) * num_motifs_in_pat );
	}

	/* copy old info back into real arrays */
    	for (m = 0; m < (*which_match); m++)
	{
            for (n = 0; n < num_motifs_in_pat; n++)
	    {
            	(*match_arr)[m][n] = temp_match_arr[m][n];
                (*bp_arr)[m][n] = temp_bp_arr[m][n];
                (*pcts_arr)[m][n] = temp_pcts_arr[m][n];
                (*block_nbr_arr)[m][n] = temp_block_arr[m][n];
	    }
	}

	/* initialize first columns of the new row of the real arrays */
        for (n = 0; n < num_motifs_in_pat; n++)
        {
            (*match_arr)[*which_match][n] = (*match_arr)[*which_match - 1][n];
            (*bp_arr)[*which_match][n] = (*bp_arr)[*which_match - 1][n];
            (*pcts_arr)[*which_match][n] = (*pcts_arr)[*which_match - 1][n];
            (*block_nbr_arr)[*which_match][n] = (*block_nbr_arr)[*which_match - 1][n];
        }

	/* initialize single values in real arrays */
        lowest_cutoff = (floor ((score - min_score) / ((float) (max_score - min_score)) * 100) / 100);

        (*match_arr)[*which_match][which_motif] = cur_pos;
        (*bp_arr)[*which_match][which_motif] = bp;
        (*pcts_arr)[*which_match][which_motif] = lowest_cutoff;
	(*block_nbr_arr)[*which_match][which_motif] = block_nbr;

	/* clean up temporary arrays */
    	free_int_arr( temp_match_arr, (*which_match) );
        free_int_arr( temp_bp_arr, (*which_match) );
        free_int_arr( temp_block_arr, (*which_match) );
 	free_float_arr( temp_pcts_arr, (*which_match) );

    /* first motif in first match */
    } else {

        lowest_cutoff = (floor ((score - min_score) / ((float) (max_score - min_score)) * 100) / 100);               

    	(*match_arr)[cur_match][which_motif] = cur_pos;
        (*bp_arr)[cur_match][which_motif] = bp;
        (*pcts_arr)[*which_match][which_motif] = lowest_cutoff;
	(*block_nbr_arr)[cur_match][which_motif] = block_nbr;
    }

    /* debug */ /*
    for (m = 0; m < (*which_match) + 1; m++)
        for (n = 0; n < num_motifs_in_pat; n++)
            printf("MATCH ARR [%d][%d] is %d\n", m, n, (*match_arr)[m][n]);
    */
}


/* Compute the avg pct for all motifs of each match, storing each result in pct[m][0]. */
static void get_avg_pcts( float **pcts_arr, int tot_matches, int num_motifs_in_pat )
{
    int m, n;			/* array indices */

    for (m = 0; m < tot_matches; m++)
    {
	for (n = 1; n < num_motifs_in_pat; n++)
	    pcts_arr[m][0] += pcts_arr[m][n];
	pcts_arr[m][0] /= num_motifs_in_pat;
    }
}


/* Compute the lowest pct of all motifs of each match, storing each result in pct[m][0]. */
static void get_lowest_pcts( float **pcts_arr, int tot_matches, int num_motifs_in_pat )
{ 
    int m, n;                   /* array indices */
    float lowest_pct;
        
    for (m = 0; m < tot_matches; m++) 
    {       
	lowest_pct = INT_MAX;
        for (n = 0; n < num_motifs_in_pat; n++)
            if (pcts_arr[m][n] < lowest_pct)
		lowest_pct = pcts_arr[m][n];
        pcts_arr[m][0] = lowest_pct;
    }       
}


/* Does the given match have a hit for all motifs in the pattern? */
static int valid_match( int *match, int num_motifs_in_pat )
{
    int m;		/* array index */

    for (m = 0; m < num_motifs_in_pat; m++)
	if (match[m] == -1)
	    return 0;

    return 1;
}


/* Add the given match to match info. */
static void add_to_matches( int num_motifs, int which_seq, int nrow,
			    struct MATCH **match_head, int match_len, float pct,
			    int *block_nbrs, int startpt, int endpt, int gap_in_top,
			    int report_to_top, int *bp_positions, int cur_dir )
{
    int duplicate = 0;			/* does the new match occur at the same place 
					   in another seq? (bool) */
    struct MATCH *match_step = NULL, 	/* pointer for moving through match list */
		 *new_match = NULL;	/* new match to be added to the list */

    match_step = *match_head;

    /* stop at position where the match is a duplicate, or move to the end of the list */
    while (match_step != NULL)
    {
 	if (same_match( match_step, bp_positions, num_motifs, block_nbrs ))
	{
	    duplicate = 1;
	    break;
	}
        match_step = match_step->next;
    }

    if (duplicate) {		/* its the same match, on a different sequence */
	match_step->which_seqs[which_seq] = 1;
	match_step->startpt[which_seq] = startpt;
	match_step->endpt[which_seq] = endpt;
	(match_step->num_seqs)++;
        match_step->pct[which_seq] = pct;

    } else {			/* its a new match in the list */
        new_match = ckalloc( sizeof( struct MATCH ) );

	fill_in_match_info( new_match, num_motifs, which_seq, match_len, pct, nrow,
			    block_nbrs, startpt, endpt, gap_in_top, bp_positions );
        add_to_match_list( new_match, match_head, num_motifs, nrow, report_to_top, 
			   cur_dir );
    }
}


/* Do the two matches occur at the same positions and block in the alignment? */
static int same_match( struct MATCH *match_step, int *bp_positions, int num_motifs,
		       int *block_nbrs )
{
    int i, j;		/* array indices */

    for (i = 0; i < num_motifs; i++)
    {
	if (match_step->positions[i] != bp_positions[i])
	    return 0;
	for (j = 0; j < num_motifs; j++)
	    if (match_step->block_nbrs[j] != block_nbrs[j])
	        return 0;
    }

    return 1;
}


/* Have we passed up the spot where the match should be inserted in the list? */
static int too_far_in_match_list( struct MATCH *match_step, struct MATCH *new_match, 
			 	  int num_motifs, int nrow, int reltop, int cur_dir )
{
    int i,		/* array index */
	match_row = 0,	/* row the hit occurs on in the match already in the list */
	new_row = 0;	/* row the hit occurs on in the new match to be added */

    /* first, sort by topmost sequence the match occurs in if not reported relative
       to topmost sequence */
    if (!reltop)
	for (i = 0; i < nrow; i++)
	{
	    if (match_step->which_seqs[i] && !new_match->which_seqs[i])
	        return 0;
            if (!match_step->which_seqs[i] && new_match->which_seqs[i])
	        return 1;
	    if (match_step->which_seqs[i] && new_match->which_seqs[i])
		break;
	}

    /* determine row that each occurs on */
    for (i = 0; i < nrow; i++)
    {
    	if (match_step->which_seqs[i])
	    match_row = i;
	if (new_match->which_seqs[i])
	    new_row = i;
    }

    /* compare positions of these rows */
    if (match_step->startpt[match_row] > new_match->startpt[new_row])
    	return 1;
    else
    	return 0;
}


/* Add the given match structure to the match list. */
static void add_to_match_list( struct MATCH *new_match, struct MATCH **match_head, 
			       int num_motifs, int nrow, int report_to_top,
			       int cur_dir )
{
    struct MATCH *match_step,		/* pointer for moving through match list */
		 *prev_step = NULL;	/* trailing pointer for moving through list */

    match_step = *match_head;
    if (match_step == NULL)
	*match_head = new_match;
    else
    {
	prev_step = *match_head;
	while (match_step != NULL)
	{
	    /* only possible to record a match out of order if hits are reported in
	       relation to topmost sequence or we are on the reverse strand */
	    if ((report_to_top || cur_dir == REVERSE) && 
		 too_far_in_match_list( match_step, new_match, num_motifs, nrow,
				       	report_to_top, cur_dir ))
		break;
	    prev_step = match_step;
	    match_step = match_step->next; 
	}
	new_match->next = match_step;
	if (prev_step == match_step)		/* new header */
	    *match_head = new_match;
	else					
	    prev_step->next = new_match;
    }
}


/* Fill match structure with given information. */
static void fill_in_match_info( struct MATCH *new_match, int num_motifs, int which_seq,
				int match_len, float pct, int nrow, int *block_nbrs,
				int startpt, int endpt, int gap_in_top, int *bp_positions)
{
    int i;		/* array index */

    /* allocate memory */
    new_match->positions = ckalloc( sizeof( int ) * num_motifs );
    new_match->which_seqs = ckalloc( sizeof( int ) * nrow );
    new_match->pct = ckalloc( sizeof( float ) * nrow );
    new_match->startpt = ckalloc( sizeof( int ) * nrow );
    new_match->endpt = ckalloc( sizeof( int ) * nrow );
    new_match->block_nbrs = ckalloc( sizeof( int ) * num_motifs );

    /* copy information */
    for (i = 0; i < num_motifs; i++)
        new_match->positions[i] = bp_positions[i];
    for (i = 0; i < nrow; i++)
    {
        new_match->which_seqs[i] = 0;
	new_match->startpt[i] = -1;
	new_match->endpt[i] = -1;
        new_match->pct[i] = INT_MIN;
    }
    new_match->which_seqs[which_seq] = 1;
    new_match->startpt[which_seq] = startpt;
    new_match->endpt[which_seq] = endpt;
    for (i = 0; i < num_motifs; i++)
    	new_match->block_nbrs[i] = block_nbrs[i];
    new_match->num_seqs = 1;
    new_match->match_len = match_len;
    new_match->gap_in_top = gap_in_top;
    new_match->pct[which_seq] = pct;
    new_match->next = NULL;
}


/* Returns 1 if the given hit falls within an exons or gene region. */
static int within_exon( int startpt, int endpt, int *beg, int *end, int *dir, int cur_dir,
			int num_to_mask )
{
    int which_exon;		/* array index of the current exon we are on */

    for (which_exon = 0; which_exon < num_to_mask; which_exon++)
	if (cur_dir == dir[which_exon])
	{
	    /* starts within exon region */
	    if ((startpt >= beg[which_exon]) && (startpt <= end[which_exon]))
		return 1;

	    /* ends within exon region */
	    if ((endpt >= beg[which_exon]) && (endpt <= end[which_exon]))
	    	return 1;

	    /* runs through exon region */
	    if ((startpt <= beg[which_exon]) && (endpt >= end[which_exon]))
		return 1;
	}

    /* if we have made it this far, it is not in the range */
    return 0;
}


/* Adjust range for possible gaps in sequence. */
static void adjust_from_and_to( int *from, int *to, int *col2pos, int ncol, int cur_dir,
				char align_type, int max_pat_len )
{
    int j = 0, 			/* array index */
	temp = 0;		/* temporary storage of from */

    /* dont need to worry about gaps in FASTA files */
    if (align_type == 'f') {
	if (cur_dir == FORWARD) {
	    (*from)--;
	    (*to)--;
	} else {
            temp = *from;
            *from = ncol - (*to + 1) - (max_pat_len - 1);
            *to = ncol - (temp + 1);
	}

    /* adjust for gaps in multiple alignments */
    } else if (align_type == 'm') {
	/* for forward strand, adjust user's input 'from' value (which is relative to
	   the topmost strand) to the column to start the search on in aligment. */
	if (cur_dir == FORWARD)
	{
    	    for (j = *from - 1; j < ncol; j++)
    	    {
		/* don't need to calculate from range if starting at beginning */
            	if ((*from != 1) && (col2pos[j] == *from))
            	    temp = j;
                if (col2pos[j] == *to)
            	    break;
            }
            *from = temp;
            *to = j;

	/* for the reverse strand, use the values obtained for the forward strand
	   and reverse them, making them relative to the reverse strand.  the 
	   beginning of the range (from) also needs to have the maximum possible
	   length of the pattern subtracted from it, so that it starts searching
	   properly from the end of the pattern.  since not all patterns will
	   be the maximum possible pattern length, this may lead to a few extra
	   hits not within the range.  as such, these are filtered out in the
	   function 'output_match' */
        } else {
	    temp = *from;
	    *from = ncol - (*to + 1) - (max_pat_len - 1);
	    *to = ncol - (temp + 1);
    	}
    } else {
	/* do nothing extra for type 'b' files */
    }
    check_boundaries( from, to, 0, ncol );
}


/* Get exon option and file from user, call get_exons. */
static int get_exon_info( int **beg, int **end, int **dir, int *exon_seq, int nrow,
			  char **exon_file )
{
    char mask_type, 		/* type of masking to be performed (g, c, e) */
         *long_string;  	/* utility string for reading in user input */
    int exlen = 0, 		/* number of regions to be masked out */
	valid_choice = 0;	/* is the choice entered by the user valid? (bool) */
    FILE *fp = NULL;		/* file pointer */

    long_string = ckalloc( STR_SIZE );

    fprintf(SCREEN, "\n\nExon options:\n");
    fprintf(SCREEN, "\n    g...mask out entire genes ('<' line in exons file.)\n");
    fprintf(SCREEN, "    c...mask out coding region of genes ('+' line in exons file.)");
    fprintf(SCREEN, "\n    e...mask out each individual exon");
    fprintf(SCREEN, " (unlabeled lines in exons file.)\n\n");
    fprintf(SCREEN, CHOICE_PROMPT);

    while (!valid_choice)
    {
        mask_type = get_menu_choice();
	if (mask_type == 'g' || mask_type == 'G' ||
	    mask_type == 'c' || mask_type == 'C' ||
	    mask_type == 'e' || mask_type == 'E')
		valid_choice = 1;
	else
            print_invalid_menu_choice( mask_type );
    }

    fprintf(SCREEN, "\n\nWhich sequence in the alignment are the exon regions located?");
    fprintf(SCREEN, "\n(in the range 0 to %d.)  ", nrow - 1);
    get_entry (&long_string);
    *exon_seq = atoi( long_string );
    while ((*exon_seq < 0) || (*exon_seq >= nrow))
    {
        sprintf( err_string, "Value entered must be between 0 and %d.", (nrow - 1) );
        error( err_string, "", NOKILL );
	fprintf(SCREEN, "Please enter a new value.  ");
	get_entry (&long_string);
    	*exon_seq = atoi( long_string );
    }

    fprintf(SCREEN, "\n\nEnter file name for exons file.  ");
    get_entry (exon_file);
    fp = ckopen_nk( *exon_file, "r" );
    while (exlen == 0)
    {
	if (fp == NULL)
	{
            error( "Invalid file name or format.", "", NOKILL );
            fprintf(SCREEN, "Please enter a valid file name and format.  ");
	    get_entry (exon_file);
            fp = ckopen_nk( *exon_file, "r" );
	    exlen = 0;
	} else {
            exlen = get_exons( fp, beg, end, dir, mask_type );
	    if (exlen == 0)
	    {
		if (fclose( fp ) != 0)
		{
		    sprintf( err_string, "Could not close '%s'.", *exon_file );
		    error( err_string, "", NOKILL );
		}
		fp = NULL;
	    }
	}
    }

    ckfree( long_string );
    return (exlen);
}


/* read exons from exon file into 3 parallel arrays */
static int get_exons( FILE *fp, int **Beg, int **End, int **Dir, int mask_type )
{
    char ch, 				/* single char from buffer */
         *buf,		          	/* buffer read in from one line of exon file */
         *long_string;          	/* utility string for reading input from buffer */
    int cur_dir = FORWARD, 		/* current direction on strand */
	num_to_mask = 0, 		/* number of genes to mask */
	int1 = 0, int2 = 0, 		/* ints read in from buffer */
	backup_int1 = 0, 		/* previous values of int1 and int2 */
	backup_int2 = 0,
	i = 0, 				/* single int value */
	*beg, *end, *dir;		/* parallel arrs with beg, end and dir of region*/

    /* allocate memory */
    buf = ckalloc( BUFSIZE );
    long_string = ckalloc( STR_SIZE );

    /* count number of exons or genes in file */
    switch (mask_type)
    {
        case 'g' : case 'G' :
        case 'c' : case 'C' :

            while (fgets( buf, BUFSIZE, fp ))			      /* genes */
            	if((sscanf( buf, " %1[><|] %d %d %s", &ch, &int1, &int2, 
						      long_string )) == 4)
		    num_to_mask++;
	    break;

        case 'e' : case 'E' :

    	    while (fgets( buf, BUFSIZE, fp ))
            	if ((sscanf( buf, " %d %d", &int1, &int2 )) == 2)     /* exons */
	    	    num_to_mask++;
	    break;

	default :
	    error( "Internal 'mask_type' error.", "", KILL );
    }

    /* allocate memory for parallel arrays */
    *Beg = beg = (int *) ckalloc( num_to_mask * sizeof( int ) );
    *End = end = (int *) ckalloc( num_to_mask * sizeof( int ) );
    *Dir = dir = (int *) ckalloc( num_to_mask * sizeof( int ) );

    /* skip optional title line if its there */
    rewind( fp );
    fgets( buf, BUFSIZE, fp );
    if (starts( skip_ws( buf ), ">" ) || starts( skip_ws( buf ), "<" ) || 
	starts( skip_ws( buf ), "|" ))
	    rewind( fp );

    while (fgets( buf, BUFSIZE, fp ))
    {
	if (starts( skip_ws( buf ), "#" ))				/* comment line */
	    continue;
        else if (starts( skip_ws( buf ), ">" ) || starts( skip_ws( buf ), "<" ) || 
		 starts( skip_ws( buf ), "|" ))
	{
	    /* '|' is treated as '>' */
            cur_dir = (starts( skip_ws( buf ), "<" ) ? REVERSE : FORWARD);
	    if ((mask_type == 'g') || (mask_type == 'G'))
	    {
            	if (sscanf(buf, " %1[>|<] %d %d %s", &ch, &int1, &int2, long_string) == 4)
		{
		    check_exon_positions( int1, int2, buf );
                    update_exon_info( beg, end, dir, &i, int1, int2, cur_dir );
		}
		else
		    error( "Invalid exons file format.", buf, KILL );
	    }
	    else if ((mask_type == 'c') || (mask_type == 'C'))
	    {
                if (sscanf(buf, " %1[>|<] %d %d %s", &ch, &int1, &int2, long_string) == 4)
	  	{
                    check_exon_positions( int1, int2, buf );
		    backup_int1 = int1;
		    backup_int2 = int2;
		}
		else
                    error( "Invalid exons file format.", buf, KILL );

		fgets( buf, BUFSIZE, fp );

		/* if there is a '+' line, use those values
		   otherwise, use the values from the previous gene line */
		if (sscanf( buf, " %1[+] %d %d", &ch, &int1, &int2 ) == 3)
		{
                    check_exon_positions( int1, int2, buf );
		    if ((int1 < backup_int1) || (int2 > backup_int2))
			error( "Coding region does not fall within gene boundaries.",
			       buf, KILL );
                    update_exon_info( beg, end, dir, &i, int1, int2, cur_dir );

		}
	   	else 
                    update_exon_info( beg, end, dir, &i, backup_int1, backup_int2, 
				       cur_dir );
	    }
	}
        else if (sscanf( buf, " %d %d", &int1, &int2 ) == 2)	/* exon line */
	{
	    check_exon_positions( int1, int2, buf );
            if ((mask_type == 'e') || (mask_type == 'E'))
                update_exon_info( beg, end, dir, &i, int1, int2, cur_dir );
	    else
		continue;
	}
	else							/* other line */
	    continue;
    }

    /* free allocated memory */
    ckfree( buf );
    ckfree( long_string );

    return (num_to_mask);
}


/* Check if the given exon positions are valid */
static void check_exon_positions( int int1, int int2, char *buf )
{
    if (int1 < 1 || int2 < 1)
    	error( "Non-positive positions in exons file.", buf, KILL );
    if (int1 > int2)
	error( "Positions in non-increasing order.", buf, KILL );
}


/* Update exon arrays and index variable i. */
static void update_exon_info( int *beg, int *end, int *dir, int *i, int num1, int num2,
			      int cur_dir )
{
    beg[*i] = num1;
    end[*i] = num2;
    dir[*i] = cur_dir;
    (*i)++;
}


/* Compute the score for the given motif at the given position in the alignment. */
static float motif_score( char *seq, struct ITEM *step, int cp, int ncol, Block_ptr b_ptr,
			  Block_ptr block_head, Contig_ptr contig, int which_row,
			  int cur_dir, int count_lc )
{
    char *rev_motif_seq = NULL;	/* seq to be searched for the motif, rev strand only */
    int pos;			/* current position in the sequence */
    float tot_score = 0;	/* running score of the motif */

    /* if there isn't enough room for the motif (runs off the edge), return -infinity */
    if ((!b_ptr) && (cp + step->length > ncol))
	return INT_MIN;

    /* search on reverse strand */
    if (step->compl)
    {
	rev_motif_seq = ckalloc( step->length * sizeof( char ) );
        for (pos = 0; pos < step->length; pos++)
	{
            if ((b_ptr && ((pos + cp) >= b_ptr->ncol)))
	    {
                move_to_adjacent_block( &b_ptr, &ncol, block_head, contig, &which_row,
                                        cur_dir );
                if (b_ptr == NULL)
                    return (INT_MIN);
		cp = -pos;
	    }
            rev_motif_seq[pos] = (b_ptr ? b_ptr->seg[which_row].chars[pos + cp] 
					: seq[pos + cp]);
	}
        do_revcomp( rev_motif_seq, step->length );
        for (pos = 0; pos < step->length; pos++)
            tot_score += get_score( step, rev_motif_seq[pos], pos + 1, count_lc );

    	ckfree( rev_motif_seq );
    } else {
	for (pos = 0; pos < step->length; pos++)
	{
	    if ((b_ptr && (pos + cp) >= b_ptr->ncol))
	    {
                move_to_adjacent_block( &b_ptr, &ncol, block_head, contig, &which_row,
                                    	cur_dir );
		if (b_ptr == NULL)
		    return (INT_MIN);
		cp = -pos;
	    }
	    tot_score += get_score( step, (b_ptr ? b_ptr->seg[which_row].chars[pos + cp]
						 : seq[pos + cp]), pos + 1, count_lc );
	}
    }
    return (tot_score);
}


/* Given the sequence char and motif we are currently on, return corresponding score. */
static float get_score( struct ITEM *step, char seq_char, int pos, int count_lc )
{
    if (seq_char == 'A') return (step->submat[rowA][pos]);
    if (seq_char == 'C') return (step->submat[rowC][pos]);
    if (seq_char == 'G') return (step->submat[rowG][pos]);
    if (seq_char == 'T') return (step->submat[rowT][pos]);

    /* treat lowercase nucleotides as if they were uppercase */
    if (count_lc)
    {
    	if (seq_char == 'a') return (step->submat[rowA][pos]);
    	if (seq_char == 'c') return (step->submat[rowC][pos]);
    	if (seq_char == 'g') return (step->submat[rowG][pos]);
    	if (seq_char == 't') return (step->submat[rowT][pos]);
    }

    return (INT_MIN); 
}


/* Given the column in the alignment array, return the start and end of the match
   relative to the proper sequence in the alignment. */
static void get_endpoints( int pos, int dir, int match_len, int ncol, int *col2pos,
		           int *start, int *end, int *gap_in_top, char align_type,
			   Block_ptr block, Block_ptr block_head, Contig_ptr contig,
			   int which_row, int report_to_top )
{
    if (align_type == 'b')
    {
	if (dir == REVERSE)
	    get_ends_b_rev( pos, match_len, start, end, gap_in_top, block, block_head,
			    contig, which_row, report_to_top );
	else
            get_ends_b_for( pos, match_len, start, end, gap_in_top, block, block_head,
                            contig, which_row, report_to_top );

    } else {
        if (dir == REVERSE)
            get_ends_m_rev( pos, match_len, ncol, col2pos, start, end, gap_in_top,
		            report_to_top );
        else
            get_ends_m_for( pos, match_len, ncol, col2pos, start, end, gap_in_top,
			    report_to_top );
    }
}


/* Get the endpoints on the reverse strand for a type 'b' alignment file. */
static void get_ends_b_rev( int pos, int match_len, int *start, int *end,
			    int *gap_in_top, Block_ptr block, Block_ptr block_head,
			    Contig_ptr contig, int which_row, int report_to_top )
{
    char *p,		/* pointer to move around sequence */
	 *sp;		/* pointer to beginning of sequence */
    int k, 		/* array index */
        bp,             /* position relative to current block */
        which_seq,      /* which seq we are currently dealing with */
        lowest_pos,     /* lowest possible position for output */
        highest_pos,    /* highest possible position for output */
	changed_blocks,	/* did we change to an adjacent block? */
	old_start, 	/* original value of start (before decrementing it due to gaps) */
        adjust = 0;     /* used to find end col of match, adjusts for changing blocks */
    Seg_ptr s;          /* ptr used for calls to find_block and adjacent_block */

    /* check if start or end align to nothing in topmost sequence */
    if (report_to_top)
	if (block->seg[0].contig_nbr != 0)
	{
	    *start = -1;
	    *end = -1;
	    *gap_in_top = 1;
	    return;
	}
	
    /* initialize vars */
    sp = block->seg[report_to_top ? 0 : which_row].chars;
    bp = block->seg[which_row].beg - 1 + block->ncol - (pos + 1);
    which_seq = block->seg[which_row].contig_nbr;
    lowest_pos = block->seg[report_to_top ? 0 : which_row].beg;
    highest_pos = block->seg[report_to_top ? 0 : which_row].end;
    old_start = pos - block->seg[which_row].beg + 1;

    /* initialize start and end */
    *end = pos;
    *start = *end - match_len + 1;
    old_start = *start;
    *gap_in_top = 0;

    /* printf("INIT S %d E %d\n", *start, *end); */ /* debug */

    /* increment end if it aligns to a gap so that it doesnt get cut off */
    if ((report_to_top) && (sp[bp] == ' '))
    {
        (*end)++;
        *gap_in_top = 1;
    }

    /* for every gap in the sequence, decrement end position of match */
    changed_blocks = 0;
    for (p = sp + bp, k = 0; k < match_len; ++k, ++p, ++bp)
    {
        if ((block) && (bp >= block->ncol))
        {
            find_block( block_head, which_seq, block->seg[which_row].beg, contig, &s );
            adjust += block->ncol;
            if ((block = prev_block( &s, contig, block_head )) != NULL)
            {
		changed_blocks = 1;
                bp = 0;
                which_row = get_row_from_seq_num( which_seq, block->nrow, block );
    		if (which_row == -1)
        	    error( "Requested seq not found in given block in get_ends_b_rev.",
               		   "", KILL );
		lowest_pos = block->seg[report_to_top ? 0 : which_row].beg;
                p = block->seg[(report_to_top ? 0 : which_row)].chars;
		sp = p;
            } else {
                error("Could not find prev block in get_ends_b_rev.", "", KILL);
	    }
        }

        if (*p == ' ')
            (*end)--;
    }

    /* analogous to col2pos lookup */
    if (!changed_blocks) 
    {
        for (p = sp + block->ncol - 1, k = 0; 
	     k < old_start - block->seg[(report_to_top ? 0 : which_row)].beg + 1; ++k, --p)
        {
    	    if (*p == ' ')
            {
                (*start)--;
                (*end)--;
            }
        }
    }

    /* decrement start if it aligns to a gap so that it doesnt get cut off */
    if ((report_to_top) &&
        (block->seg[0].chars[block->ncol - (old_start + 1) + block->seg[which_row].beg - 1]
		 == ' '))
    {
        (*start)--;
        *gap_in_top = 1;
    }

    /* get final values for start and end */
    (*start)++;
    (*end)++;
    check_boundaries( start, end, lowest_pos, highest_pos );
}


/* Get the endpoints on the forward strand for a type 'b' alignment file. */
static void get_ends_b_for( int pos, int match_len, int *start, int *end,
			    int *gap_in_top, Block_ptr block, Block_ptr block_head,
			    Contig_ptr contig, int which_row, int report_to_top )
{
    char *p,            /* pointer to move around sequence */
         *sp;           /* pointer to beginning of sequence */
    int k,              /* array index */
	bp,		/* position relative to current block */
	which_seq, 	/* which seq we are currently dealing with */
	lowest_pos,	/* lowest possible position for output */
	highest_pos,	/* highest possible position for output */	
	start_pos,	/* starting position in the block */
	adjust = 0;	/* used to find end col of match, adjusts for changing blocks */
    Seg_ptr s;          /* ptr used for calls to find_block and adjacent_block */

    /* check if start or end align to nothing in topmost sequence */
    if (report_to_top)
    	if (block->seg[0].contig_nbr != 0)
	{
    	    *start = -1;
	    *end = -1;
	    *gap_in_top = 1;
	    return;
	}

    /* initialize vars */
    sp = block->seg[report_to_top ? 0 : which_row].chars;
    bp = pos - block->seg[which_row].beg + 1;
    which_seq = block->seg[which_row].contig_nbr;
    lowest_pos = block->seg[report_to_top ? 0 : which_row].beg;
    highest_pos = block->seg[report_to_top ? 0 : which_row].end;
    start_pos = pos - block->seg[which_row].beg + 1;

    /* initialize start and end */
    *start = start_pos;
    *end = start_pos + match_len - 1;
    *gap_in_top = 0;
    /* printf("INIT S %d E %d\n", *start, *end); */ /* debug */

    /* analogous to col2pos lookup */
    for (p = sp, k = 0; k < start_pos; ++k, ++p)
        if (*p == '-')
        {
            (*start)--;
            (*end)--;
        }
    /* printf("AFTER COL2POS S %d E %d\n", *start, *end); */ /* debug */

    /* decrement start if it aligns to a gap so that it doesnt get cut off */
    if ((report_to_top) && (sp[start_pos] == '-'))
    {
        (*start)--;
        *gap_in_top = 1;
    }

    /* for every gap in the sequence, decrement end position of match */
    for (p = sp + bp, k = 0; k < match_len; ++k, ++p, ++bp)
    {
    	if ((block) && (bp /*- block->seg[which_row].beg + 1 */ >= block->ncol))
        {
            find_block( block_head, which_seq, block->seg[which_row].beg, contig, &s );
	    adjust += block->ncol;
            if ((block = next_block( &s, contig, block_head )) != NULL)
            {
		bp = 0;
		which_row = get_row_from_seq_num( which_seq, block->nrow, block );
                if (which_row == -1)
                    error( "Requested seq not found in given block in get_ends_b_for.",
                           "", KILL );
		highest_pos = block->seg[report_to_top ? 0 : which_row].end;
            	p = block->seg[(report_to_top ? 0 : which_row)].chars;
            } else {
                error("Could not find next block in get_ends_b_for.", "", KILL);
            }
	}
        if (*p == '-')
            (*end)--;
    }

    /* printf("AFTER DECS, %d %d\n", *start, *end); */ /* debug */

    /* increment end if it aligns to a gap so that it doesnt get cut off */
    if ((report_to_top) && 
	(block->seg[0].chars[start_pos + match_len - 1 - adjust] == '-'))
    {
	(*end)++;
 	*gap_in_top = 1;
    }

    /* get final values for start and end */
    (*start)++;
    (*end)++;
    (*start) += (lowest_pos - 1);
    (*end) += (lowest_pos - 1);
    check_boundaries( start, end, lowest_pos, highest_pos );
}


/* Get the endpoints on the reverse strand for a type 'm' alignment file. */
static void get_ends_m_rev( int pos, int match_len, int ncol, int *col2pos, int *start, 
			    int *end, int *gap_in_top, int report_to_top )
{
    int k,              /* array index */
	highest_pos;	/* highest possible position for output */

    /* initialize start and end, make pos relative to reverse strand */
    pos = ncol - pos - 1;
    *end = col2pos[pos];
    *start = col2pos[pos - (match_len - 1)];
    *gap_in_top = 0;

    /* if the positions are in relation to the topmost sequence */
    if (report_to_top)
    {
    	/* determine highest possible position */
    	for (k = ncol - 1; col2pos[k] == 0; k--);
    	highest_pos = col2pos[k];

        /* if start aligns to a gap,
	   decrement start for each gap so the start doesnt get cut off */
        if (*start == 0)
        {
	    for (k = pos - (match_len - 1); (col2pos[k] == 0) && (k >= 0); k--);
	    *start = col2pos[k];
	    *gap_in_top = 1;
	}

        /* if end aligns to a gap,
	   increment end for each gap so the end doesnt get cut off */
	if (*end == 0)
	{
            for (k = pos; (col2pos[k] == 0) && (col2pos[k] <= highest_pos); k++);
	    *end = col2pos[k];
	    *gap_in_top = 1;
	}

        /* dont let start and end go beyond their boundaries in sequence 0 */
        check_boundaries( start, end, 1, highest_pos );
    }
}


/* Get the endpoints on the forward strand for a type 'm' alignment file. */
static void get_ends_m_for( int pos, int match_len, int ncol, int *col2pos, int *start,
			    int *end, int *gap_in_top, int report_to_top )
{
    int k,              /* array index */
        highest_pos;    /* highest possible position for output */

    /* initialize start and end */
    *start = col2pos[pos];
    *end = col2pos[pos + (match_len - 1)];
    *gap_in_top = 0;

    /* if the positions are in relation to the topmost sequence */
    if (report_to_top)
    {
    	/* determine highest possible position */
	for (k = ncol - 1; col2pos[k] == 0; k--);
	highest_pos = col2pos[k];

        /* if start aligns to a gap,
	   decrement start for each gap so the start doesnt get cut off */
	if (*start == 0)
	{
            for (k = pos; (col2pos[k] == 0) && (k >= 0); k--);
            *start = col2pos[k];
	    *gap_in_top = 1;
	}

        /* if end aligns to a gap, 
	   increment end for each gap so the end doesnt get cut off */
	if (*end == 0)
	{
            for (k = pos + (match_len - 1); (col2pos[k] == 0) && 
					    (col2pos[k] <= highest_pos); k++);
            *end = col2pos[k];
	    *gap_in_top = 1;
	}

        /* dont let start and end go beyond their boundaries in sequence 0 */
        check_boundaries( start, end, 1, highest_pos );
    }	 
}


/* Adjust start and end so that they stay within the given boundaries. */
static void check_boundaries( int *start, int *end, int lowest_pos, int highest_pos )
{
    if (*end > highest_pos)
        *end = highest_pos;
    if (*start < lowest_pos)
        *start = lowest_pos;
    if (*start > *end)
	*end = *start;
}


/* Step through match list and output matches that meet the requirements */
static void output_matches( struct MATCH *match_head, int min_matches, int filter_exons,
			    int *num_filtered_exons, int *tot_matches, char **A, int nrow,
			    int cur_dir, int num_to_mask, int *beg, int *end, int *dir,
			    int old_from, int old_to, char align_type,
			    Block_ptr block_head, Contig_ptr contig, int num_motifs,
			    int exon_seq, int out_format )
{
    int which_block = -1, 		/* used to move to the proper block in the list */
	start_block = -1;
    struct MATCH *match_step = NULL;	/* pointer to move through match list */
    Block_ptr b_ptr = NULL;		/* pointer to current block */

    for (match_step = match_head; match_step; match_step = match_step->next)
	if (match_step->num_seqs >= min_matches)
	{
	    /* move b_ptr to proper position in block list */
	    if (align_type == 'b')
	    {
	    	start_block = match_step->block_nbrs[0];
            	b_ptr = block_head;
            	for (which_block = 1; which_block < start_block; which_block++)
                    b_ptr = b_ptr->next;
	    }

	    (*tot_matches) += output_match( A, match_step, nrow, cur_dir, old_from,
				  	    old_to, align_type, b_ptr, block_head, contig,
					    filter_exons, beg, end, dir, num_to_mask,
					    num_filtered_exons, num_motifs, exon_seq,
					    out_format );
	}
}


/* Output a match in the proper format. */
static int output_match( char **A, struct MATCH *match_step, int nrow, int cur_dir,
			 int old_from, int old_to, char align_type, Block_ptr block,
			 Block_ptr block_head, Contig_ptr contig, int filter_exons,
			 int *beg, int *end, int *dir, int num_to_mask, 
			 int *num_filtered_exons, int num_motifs, int exon_seq,
			 int out_format )
{
    char *p;            	/* pointer to move around sequence */
    int i, k, 			/* array indices */
	row = -1, 		/* row of the current sequence in the current block */
        bp = INT_MIN;   	/* position relative to current block */
    Seg_ptr s = NULL;          	/* ptr used for calls to find_block and adjacent_block */
    Block_ptr orig_block = NULL;/* initial block we are outputting */

    orig_block = block;

    for (i = 0; i < nrow; ++i)
	if (match_step->which_seqs[i])
	{
            /* do not count hits not within the range */
	    for (k = 0; k < nrow; k++)
		if (match_step->startpt[k] != -1)
		    break;
            if ((match_step->startpt[k] > old_to) || (match_step->startpt[k] < old_from))
	    	return 0;

	    /* filter out hits within exon regions */
    	    if (filter_exons && within_exon( match_step->startpt[exon_seq], 
					     match_step->endpt[exon_seq], beg, end, dir,
					     cur_dir, num_to_mask ))
    	    {
        	(*num_filtered_exons)++;
        	return 0;
    	    }

            /* reset block to initial place in block list */
	    block = orig_block;

 	    if (out_format == TFFIND_OUT) {
            	printf("%-4d%c%-8d %-8d", i, match_step->gap_in_top ? '!' : ' ', 
		   	(match_step->startpt[i]), (match_step->endpt[i]));
            	if (align_type == 'b')
		    for (k = 0; k < num_motifs; k++)
                    	printf("%-5d  ", match_step->block_nbrs[k]);

            	printf(" %-7c %-6.2f ", (cur_dir == REVERSE ? '-' : '+'), 
			match_step->pct[i]);

	    } else if (out_format == GFF_OUT) {
		printf("%d	tffind	tfbs	%d	%d	%.2f	%c	.	",
			i, match_step->startpt[i], match_step->endpt[i], 
			match_step->pct[i], (cur_dir == REVERSE ? '-' : '+'));

	    } else {
        	error( "Internal error: unknown output format", "", KILL );
	    }

	    if (align_type == 'b')
	    {
                row = get_row_from_seq_num( i, block->nrow, block );
                if (row == -1)
                    error( "Requested seq not found in block in output_match.",
                           "", KILL );
		bp = match_step->positions[0];
		p = block->seg[row].chars + bp;
 	    } else
		p = A[i] + match_step->positions[0];

            for (k = 0; k < match_step->match_len; ++k, ++p, ++bp)
	    {
		if ((block) && (bp >= block->ncol))    /* ran out of room in this block */
		{
                    find_block( block_head, i, block->seg[row].beg, contig, &s );
		    block = adjacent_block( &s, contig, block_head, cur_dir );
		    if (block != NULL)
		    {
			row = get_row_from_seq_num( i, block->nrow, block );
                	if (row == -1)
                    	    error( "Requested seq not found in block in output_match.",
                           	   "", KILL );
			p = block->seg[row].chars;
			bp = 0;
		    }
		    else
			error("Could not find adjacent block in output_matches.","",KILL);
		}
		putchar( *p == ' ' ? '-' : *p );
	    }
            putchar('\n');
	}
    putchar('\n');

    return 1;
}


/* Print all matches in match list when in debugging mode. */
static void print_matches( struct MATCH *match_head, int num_motifs, int nrow )
{
    int i;			/* array index */
    struct MATCH *match;	/* pointer to step through match list */

    for (match = match_head; match != NULL; match = match->next)
    {
    	printf("\nBP POSITIONS: ");
    	for (i = 0; i < num_motifs; i++)
    	    printf("%d  ", match->positions[i]);
        printf("\nWHICH SEQS: ");
        for (i = 0; i < nrow; i++)
            printf("%d  ", match->which_seqs[i]);
        printf("\nSTART AND END: ");
        for (i = 0; i < nrow; i++)
            printf("S %d  E %d  |", match->startpt[i], match->endpt[i]);
        printf("\nNUM SEQS %d", match->num_seqs);
        printf("\nMATCH LEN %d", match->match_len);
        for (i = 0; i < num_motifs; i++)
            printf("\nBLOCK NBR %d", match->block_nbrs[i]);
        printf("\nPCTS: ");
        for (i = 0; i < nrow; i++)
            printf("%f  ", match->pct[i]);
        printf("\n\n");
    }
}


/* Print trailer to STDOUT. */
static void print_trailer( int tot_matches, int filter_exons, int num_filtered_exons,
			   int nrow, char **seq_names, int report_to_top, char align_type,
			   int out_format )
{
    int i;		/*array index */

    printf("\n");
    if (out_format == GFF_OUT)
        printf ("# ");

    printf("Total matches found : %d\n", tot_matches);
    if (filter_exons)
    {
    	if (out_format == GFF_OUT)
            printf ("# ");
        printf("Total hits filtered out : %d\n", num_filtered_exons);
    }

    if (tot_matches > 0)
    {
    	if (out_format == GFF_OUT)
            printf ("# ");

	if (report_to_top)
	{
    	    printf("All positions relative to sequence 0.\n");
    	    if (out_format == GFF_OUT)
        	printf ("# ");
	    printf("Positions marked with an '!'");
    	    printf(" align to a gap in sequence 0.\n");
	    if (align_type == 'b')
	    {
		if (out_format == GFF_OUT)
        	    printf ("# ");
	    	printf("Positions marked with a '-1' lie in a block that does not ");
	    	printf("contain the top-most sequence.\n");
	    }
	} else {
	    printf("All positions relative to the sequence that the hit occurs on.\n");
	}
    	printf("\n");
    	if (out_format == GFF_OUT)
            printf ("# ");

	printf("Sequence key:\n\n");
    	for (i = 0; i < nrow; i++)
	{
            if (out_format == GFF_OUT)
            	printf ("# ");
	    printf("  %d.....%s\n", i, seq_names[i]);
	}
    }
    printf("\n");
}


/* Create an array that returns the position in the sequence when given the column. */
static void get_position_in_seq( char **A, int ncol, int which_seq, int **col2pos )
{
    int j, 		/* array index */
	cp; 		/* current position relative to the current sequence */

    for (j = cp = 0; j < ncol; ++j)
    	(*col2pos)[j] = (isalpha((unsigned char)A[which_seq][j]) ? ++cp : 0);
}

/*======================================================================================*/
